# libraries
import logging
from datetime import timedelta
from pathlib import Path
import warnings
import numpy as np
import pandas as pd
import sqlalchemy as sa
from sqlalchemy import text as sqltxt
from functools import cached_property
import rasterio as rio
import rasterio.mask
import pyproj
from shapely.ops import transform as shp_transform
from shapely import distance, Point

from ..db.connections import db_engine
from ..utils.dwd import  dwd_id_to_str
from ..utils.TimestampPeriod import TimestampPeriod
from ..utils.geometry import polar_line, raster2points
from ..config import config
from ..db.models import MetaP
from .StationBases import StationPBase
from .constants import MIN_TSTP
from .StationPD import StationPD
from .StationT import StationT

# set settings
# ############
__all__ = ["StationP"]
log = logging.getLogger(__name__)

# variables
RICHTER_CLASSES = {
    "no-protection": {
        "min_horizon": 0,
        "max_horizon": 3
    },
    "little-protection": {
        "min_horizon": 3,
        "max_horizon": 7
    },
    "protected": {
        "min_horizon": 7,
        "max_horizon": 12
    },
    "heavy-protection": {
        "min_horizon": 12,
        "max_horizon": np.inf
    },
}

# class definition
##################
class StationP(StationPBase):
    """A class to work with and download 10 minutes precipitation data for one station."""
    # common settings
    _MetaModel = MetaP
    _para = "p"
    _para_base = _para
    _para_long = "Precipitation"
    _unit = "mm/10min"
    _valid_kinds = {"raw", "qn", "qc", "corr", "filled", "filled_by"}
    _best_kind = "corr"

    # cdc dwd parameters
    _ftp_folder_base = [
        "climate_environment/CDC/observations_germany/climate/10_minutes/precipitation/"]
    _cdc_col_names_imp = ["RWS_10", "QN"]
    _db_col_names_imp = ["raw", "qn"]

    # timestamp configurations
    _tstp_format_db = "%Y%m%d %H:%M"
    _tstp_dtype = "timestamp"
    _interval = "10 min"

    # aggregation
    _min_agg_to = "10 min"

    def __init__(self, id, **kwargs):
        super().__init__(id, **kwargs)
        self.id_str = dwd_id_to_str(id)

    def _get_sql_new_qc(self, period):
        # create sql_format_dict
        sql_format_dict = dict(
            para=self._para, stid=self.id, para_long=self._para_long,
            decim=self._decimals,
            **period.get_sql_format_dict(
                format="'{}'".format(self._tstp_format_db)),
            limit=0.1*self._decimals) # don't delete values below 0.1mm/10min if they are consecutive

        # check if daily station is available
        sql_check_d = """
            SELECT EXISTS(
                SELECT *
                FROM information_schema.tables
                WHERE table_schema = 'timeseries'
                    AND table_name = '{stid}_{para}_d'
            );""".format(**sql_format_dict)
        with db_engine.connect() as con:
            daily_exists = con.execute(sqltxt(sql_check_d)).first()[0]

        # create sql for dates where the aggregated 10 minutes measurements are 0
        # althought the daily measurements are not 0
        # or where the aggregated daily sum is more than the double of the daily measurement, when the daily measurement is more than 10 mm
        if daily_exists:
            sql_dates_failed = """
                WITH ts_10min_d AS (
                    SELECT (ts.timestamp - INTERVAL '6h')::date as date, sum("raw") as raw
                    FROM timeseries."{stid}_{para}" ts
                    WHERE ts.timestamp BETWEEN {min_tstp} AND {max_tstp}
                    GROUP BY (ts.timestamp - INTERVAL '6h')::date)
                SELECT date
                FROM timeseries."{stid}_{para}_d" ts_d
                LEFT JOIN ts_10min_d ON ts_d.timestamp::date=ts_10min_d.date
                WHERE ts_d.timestamp BETWEEN {min_tstp}::date AND {max_tstp}::date
                    AND ((ts_10min_d.raw = 0 AND ts_d.raw <> 0) OR
                         (ts_10min_d.raw >= 10*{decim} AND ts_10min_d.raw >= (ts_d.raw*2)))
            """.format(**sql_format_dict)
        else:
            log.error((
                "For the {para_long} station with ID {stid} there is no timeserie with daily values. " +
                "Therefor the quality check for daily values equal to 0 is not done.\n" +
                "Please consider updating the daily stations with:\n" +
                "stats = stations.StationsPD()\n" +
                "stats.update_meta()\nstats.update_raw()"
            ).format(**sql_format_dict))
            sql_dates_failed = """
                SELECT NULL::date as date
            """

        # remove single peaks above 5mm/10min
        sql_single_peaks = """
            SELECT ts.timestamp
            FROM timeseries."{stid}_{para}" ts
            INNER JOIN timeseries."{stid}_{para}" tsb
                ON ts.timestamp = tsb.timestamp - INTERVAL '10 min'
            INNER JOIN timeseries."{stid}_{para}" tsa
                ON ts.timestamp = tsa.timestamp + INTERVAL '10 min'
            WHERE ts.raw > (5*{decim}) AND tsb.raw = 0 AND tsa.raw = 0
                AND ts.timestamp BETWEEN {min_tstp} AND {max_tstp}
        """.format(**sql_format_dict)

        # make sql for timestamps where 3 times same value in a row
        sql_tstps_failed = """
            WITH tstps_df as (
                SELECT  ts.timestamp as tstp_1,
                        ts2.timestamp as tstp_2,
                        ts3.timestamp as tstp_3
                from timeseries."{stid}_{para}" ts
                INNER JOIN timeseries."{stid}_{para}" ts2
                    on ts.timestamp = ts2.timestamp - INTERVAL '10 min'
                INNER JOIN timeseries."{stid}_{para}" ts3
                    on ts.timestamp = ts3.timestamp - INTERVAL '20 min'
                WHERE ts.qn != 3
                    AND ts.raw = ts2.raw AND ts2.raw = ts3.raw
                    AND ts.raw > {limit:n}
                    AND ts.raw is not NULL
                    AND ts.timestamp BETWEEN {min_tstp} AND {max_tstp}
            )
            SELECT tstp_1 AS timestamp FROM tstps_df
            UNION SELECT tstp_2 FROM tstps_df
            UNION SELECT tstp_3 FROM tstps_df
            UNION ({sql_single_peaks})
        """.format(sql_single_peaks=sql_single_peaks,
                   **sql_format_dict)

        # create sql for new qc values
        sql_new_qc = """
            WITH tstps_failed as ({sql_tstps_failed}),
                 dates_failed AS ({sql_dates_failed})
            SELECT ts.timestamp,
                (CASE WHEN ((ts.timestamp IN (SELECT timestamp FROM tstps_failed))
                        OR ((ts.timestamp - INTERVAL '6h')::date IN (
                            SELECT date FROM dates_failed))
                        OR ts."raw" < 0
                        OR ts."raw" >= 50*{decim})
                    THEN NULL
                    ELSE ts."raw" END) as qc
            FROM timeseries."{stid}_{para}" ts
            WHERE ts.timestamp BETWEEN {min_tstp} AND {max_tstp}
        """.format(
            sql_tstps_failed=sql_tstps_failed,
            sql_dates_failed=sql_dates_failed,
            **sql_format_dict)

        return sql_new_qc

    def _check_df_raw(self, df):
        """This function is used in the Base class on the single dataframe that is downloaded from the CDC Server before loading it in the database.

        Here the function adapts the timezone relative to the date.
        As the data on the CDC server is in MEZ before 200 and in UTC after 2000

        Some precipitation stations on the DWD CDC server have also rows outside of the normal 10 Minute frequency, e.g. 2008-09-16 01:47 for Station 662.
        Because those rows only have NAs for the measurement they are deleted."""
        # correct Timezone before 2000 -> CEWT after 2000 -> UTC
        if df.index.tzinfo is not None:
            df.index = df.index.tz_localize(None) # some are already falsy localized
        if df.index.min()>= pd.Timestamp(1999,12,31,23,0):
            df.index = df.index.tz_localize("UTC")
        elif df.index.max() < pd.Timestamp(2000,1,1,1,0):
            df.index = df.index.tz_localize("Etc/GMT-1").tz_convert("UTC")
        else:
            raise ValueError("The timezone could not get defined for the given import." + str(df.head()))

        # delete measurements outside of the 10 minutes frequency
        df = df[df.index.minute%10==0].copy()

        # check if duplicates and try to remove
        if df.index.has_duplicates:
            df = df.reset_index()\
                .groupby(self._cdc_col_names_imp + [self._cdc_date_col])\
                .first()\
                .reset_index().set_index(self._cdc_date_col)
            if df.index.has_duplicates:
                raise ValueError("There are duplicates in the DWD data that couldn't get removed.")

        # set frequency to 10 minutes
        df = df.asfreq("10min")

        # delete measurements below 0
        n_col = self._cdc_col_names_imp[self._db_col_names_imp.index("raw")]
        df.loc[df[n_col]<0, n_col] = np.nan

        return df

    @cached_property
    def _table(self):
        return sa.table(
            f"{self.id}_{self._para}",
            sa.column("timestamp", sa.DateTime),
            sa.column("raw", sa.Integer),
            sa.column("qc", sa.Integer),
            sa.column("filled", sa.Integer),
            sa.column("filled_by", sa.SmallInteger),
            sa.column("corr", sa.Integer),
            schema="timeseries")

    @db_engine.deco_create_privilege
    def _create_timeseries_table(self):
        """Create the timeseries table in the DB if it is not yet existing."""
        sql_add_table = '''
            CREATE TABLE IF NOT EXISTS timeseries."{stid}_{para}"  (
                timestamp timestamp PRIMARY KEY,
                raw int4,
                qn smallint,
                qc  int4,
                filled int4,
                filled_by int2,
                corr int4
            );
        '''.format(stid=self.id, para=self._para)
        with db_engine.connect() as con:
            con.execute(sqltxt(sql_add_table))
            con.commit()

    @staticmethod
    def _check_period_extra(period):
        """Additional checks on period used in StationBase class _check_period method."""
        # add time to period if given as date
        return period.expand_to_timestamp()

    @staticmethod
    def _richter_class_from_horizon(horizon):
        richter_class = None
        for key in RICHTER_CLASSES:
            if (horizon >= RICHTER_CLASSES[key]["min_horizon"]) and \
                (horizon < RICHTER_CLASSES[key]["max_horizon"]):
                richter_class = key
        return richter_class

    @db_engine.deco_update_privilege
    def update_horizon(self, skip_if_exist=True, **kwargs):
        """Update the horizon angle (Horizontabschirmung) in the meta table.

        Get new values from the raster and put in the table.

        Parameters
        ----------
        skip_if_exist : bool, optional
            Skip updating the value if there is already a value in the meta table.
            The default is True.
        **kwargs : dict, optional
            Additional keyword arguments catch all, but unused here.

        Returns
        -------
        float
            The horizon angle in degrees (Horizontabschirmung).
        """
        if skip_if_exist:
            horizon = self.get_horizon()
            if horizon is not None:
                return horizon

        # check if files are available
        dem_files = [Path(file) for file in config.getlist("data:rasters", "dems")]
        for dem_file in dem_files:
            if not dem_file.is_file():
                raise ValueError(
                    f"The DEM file was not found in the data directory under: \n{dem_file}")

        # get the horizon configurations
        radius = int(config.get("weatherdb", "horizon_radius"))
        proj_crs = pyproj.CRS.from_epsg(config.get("weatherdb", "horizon_crs"))
        if not proj_crs.is_projected:
            raise ValueError(
                "The CRS for the horizon calculation is not projected. " +
                "Please define a projected CRS in the config file under weatherdb.horizon_crs.")

        # get raster basic information
        dem_infos = {}
        for dem_file in dem_files:
            with rio.open(dem_file) as dem:
                # get station position
                stat_h, stat_tr = rasterio.mask.mask(
                    dem, [self.get_geom(crs=dem.crs)],
                    crop=True, all_touched=False,
                    nodata=np.nan)
                stat_geom = raster2points(stat_h, stat_tr, crs=dem.crs)\
                    .to_crs(proj_crs)\
                    .iloc[0].geometry
                xy = [stat_geom.x, stat_geom.y]
                stat_h = stat_h.flatten()[0]

                # get CRS transformer
                tr_to_dem = pyproj.Transformer.from_crs(proj_crs, dem.crs, always_xy=True)
                tr_from_dem = pyproj.Transformer.from_crs(dem.crs, proj_crs, always_xy=True)

                # get typical distance of raster cell points
                dem_tr = dem.transform
                p_top_left = (dem_tr.c, dem_tr.f)
                p_top_right = (dem_tr.c+dem_tr.a*dem.profile["width"], dem_tr.f)
                p_bottom_left= (dem_tr.c, dem_tr.f+dem_tr.e*dem.profile["height"])
                p_bottom_right = (p_top_right[0], p_bottom_left[1])
                dists = []
                for p in [p_top_left, p_top_right, p_bottom_left, p_bottom_right]:
                    p1 = Point(tr_from_dem.transform(*p))
                    dists.append(distance(
                        p1, Point(tr_from_dem.transform(p[0] + dem_tr.a, p[1] + dem_tr.e))
                    ))

            dem_infos[dem_file] = dict(
                stat_geom=stat_geom,
                xy=xy,
                stat_h=stat_h,
                tr_to_dem=tr_to_dem,
                tr_from_dem=tr_from_dem,
                max_dist = max(dists)*1.2
                )

        # check if first dem has hight information
        if np.isnan(dem_infos[dem_files[0]]["stat_h"]):
            raise ValueError(
                f"update_horizon(): No height was found in the first dem for {self._para_long} Station {self.id}. " +
                "Therefor the horizon angle could not get updated.")

        # get station center point
        stat_geom = self.get_geom(crs=proj_crs)
        xy = [stat_geom.x, stat_geom.y]

        # compute the horizon angle for each western angle
        hab = pd.Series(
            index=pd.Index([], name="angle", dtype=int),
            name="horizon",
            dtype=float)
        raise_hole_error = False
        for angle in range(90, 271, 3):
            missing_lines = [polar_line(xy, radius, angle)]
            dem_lines = None
            for dem_file, dem_info in dem_infos.items():
                with rio.open(dem_file) as dem:
                    missing_lines_next = []
                    for line in missing_lines:
                        # get raster points
                        line_dem = shp_transform(
                            dem_info["tr_to_dem"].transform,
                            line)
                        dem_np, dem_tr = rasterio.mask.mask(
                            dem, [line_dem],
                            crop=True, all_touched=True,
                            nodata=np.nan)
                        dem_line = raster2points(dem_np, dem_tr, crs=dem.crs).to_crs(proj_crs)

                        # calculate the distance to the stations raster cell
                        dem_line["dist"] = dem_line.distance(dem_info["stat_geom"])
                        dem_line.drop(
                            dem_line.loc[dem_line["dist"]==0].index,
                            inplace=True)
                        dem_line = dem_line.sort_values("dist").reset_index(drop=True)


                        # calculate the horizon angle
                        dem_line["horizon"] = np.degrees(np.arctan(
                                (dem_line["data"]-dem_info["stat_h"]) / dem_line["dist"]))

                        # check if parts are missing and fill
                        #####################################

                        # look for holes inside the line
                        for i in dem_line[dem_line["dist"].diff() > dem_info["max_dist"]].index:
                            print("hole inside", dem_line.loc(i), dem_line.loc(i-1))
                            missing_lines_next.append(
                                polar_line(dem_line.loc[i-1, "geometry"].coords,
                                        dem_line.loc[i, "dist"] - dem_line.loc[i-1, "dist"],
                                        angle))

                        # merge the parts
                        dem_lines = pd.concat(
                            [dem_lines, dem_line], ignore_index=True)

                # look for missing values at the end
                dem_max_dist = dem_lines.iloc[-1]["dist"]
                if dem_max_dist < (radius - dem_info["max_dist"]):
                    missing_lines_next.append(
                        polar_line(dem_lines.iloc[-1]["geometry"].coords[0],
                                radius - dem_max_dist,
                                angle))

                # check if no parts are missing
                if (len(missing_lines_next) == 0):
                    break
                elif dem_file == dem_files[-1]:
                    raise_hole_error = True
                else:
                    # create lines for next iteration
                    missing_lines = missing_lines_next

            hab[angle] = dem_lines["horizon"].max()

        if raise_hole_error:
            log.error(
                f"Station{self._para}({self.id}).update_horizon(): There were holes in the DEM rasters providen when calculating the horizon angle. Therefor the calculated horizon angle could be faulty.")

        # calculate the mean "horizontabschimung"
        # Richter: H’=0,15H(S-SW) +0,35H(SW-W) +0,35H(W-NW) +0, 15H(NW-N)
        horizon = max(0,
            0.15*hab[(hab.index>225) & (hab.index<=270)].mean()
            + 0.35*hab[(hab.index>=180) & (hab.index<=225)].mean()
            + 0.35*hab[(hab.index>=135) & (hab.index<180)].mean()
            + 0.15*hab[(hab.index>=90) & (hab.index<135)].mean())

        # insert to meta table in db
        self._update_meta(
            cols=["horizon"],
            values=[horizon])

        return horizon

    @db_engine.deco_update_privilege
    def update_richter_class(self, skip_if_exist=True, **kwargs):
        """Update the richter class in the meta table.

        Get new values from the raster and put in the table.

        Parameters
        ----------
        skip_if_exist : bool, optional
            Skip updating the value if there is already a value in the meta table.
            The default is True
        **kwargs : dict, optional
            Additional keyword arguments catch all, but unused here.

        Returns
        -------
        str
            The richter class name.
        """
        # check if already value in table
        if skip_if_exist:
            richter_class = self.get_richter_class()
            if self.get_richter_class() is not None:
                return richter_class

        # get the richter class
        richter_class = self._richter_class_from_horizon(
            horizon=self.update_horizon(skip_if_exist=skip_if_exist))

        # save to db
        self._update_meta(
                cols=["richter_class"],
                values=[richter_class])

        return richter_class

    @db_engine.deco_update_privilege
    def richter_correct(self, period=(None, None), **kwargs):
        """Do the richter correction on the filled data for the given period.

        Parameters
        ----------
        period : TimestampPeriod or (tuple or list of datetime.datetime or None), optional
            The minimum and maximum Timestamp for which to get the timeseries.
            If None is given, the maximum or minimal possible Timestamp is taken.
            The default is (None, None).
        **kwargs : dict, optional
            Additional keyword arguments catch all, but unused here.

        Raises
        ------
        Exception
            If no richter class was found for this station.
        """
        # check if period is given
        if not isinstance(period, TimestampPeriod):
            period = TimestampPeriod(*period)
        period_in = period.copy()
        period = self._check_period(
                period=period, kinds=["filled"])
        if not period_in.is_empty():
            sql_period_clause = \
                "WHERE timestamp BETWEEN {min_tstp} AND {max_tstp}".format(
                    **period.get_sql_format_dict(f"'{self._tstp_format_db}'"))
        else:
            sql_period_clause = ""

        # check if temperature station is filled
        stat_t = StationT(self.id)
        stat_t_period = stat_t.get_filled_period(kind="filled")
        delta = timedelta(hours=5, minutes=50)
        min_date = pd.Timestamp(MIN_TSTP).date()
        stat_t_min = stat_t_period[0].date()
        stat_t_max = stat_t_period[1].date()
        stat_p_min = (period[0] - delta).date()
        stat_p_max = (period[1] - delta).date()
        if stat_t_period.is_empty()\
                or (stat_t_min > stat_p_min
                    and not (stat_p_min < min_date)
                             and (stat_t_min == min_date)) \
                or (stat_t_max  < stat_p_max)\
                and not stat_t.is_last_imp_done(kind="filled"):
            stat_t.fillup(period=period)

        # get the richter exposition class
        richter_class = self.update_richter_class(skip_if_exist=True)
        if richter_class is None:
            raise Exception("No richter class was found for the precipitation station {stid} and therefor no richter correction was possible."\
                            .format(stid=self.id))

        # create the sql queries
        sql_format_dict = dict(
            stid=self.id,
            para=self._para,
            richter_class=richter_class,
            period_clause=sql_period_clause,
            n_decim=self._decimals,
            t_decim=stat_t._decimals
        )
        # daily precipitation
        sql_p_daily = """
            SELECT timestamp::date AS date,
                   sum("filled") AS "filled",
                   count(*) FILTER (WHERE "filled" > 0) AS "count_n"
            FROM timeseries."{stid}_{para}"
            {period_clause}
            GROUP BY timestamp::date
        """.format(**sql_format_dict)

        # add is_winter
        sql_p_daily_winter = """
            SELECT date,
			    CASE WHEN EXTRACT(MONTH FROM date) IN (1, 2, 3, 10, 11, 12)
			            THEN true::bool
			            ELSE false::bool
			    END AS is_winter,
			    "filled" AS "p_d",
			    "count_n"
			FROM ({sql_p_daily}) tsp_d
        """.format(sql_p_daily=sql_p_daily)

        # add precipitation class
        sql_p_daily_precip_class = """
            SELECT
                date, "count_n", "p_d",
                CASE WHEN (tst."filled" >= (3 * {t_decim}) AND "is_winter") THEN 'precip_winter'
                    WHEN (tst."filled" >= (3 * {t_decim}) AND NOT "is_winter") THEN 'precip_summer'
                    WHEN (tst."filled" <= (-0.7 * {t_decim})::int) THEN 'snow'
                    WHEN (tst."filled" IS NULL) THEN NULL
                    ELSE 'mix'
                END AS precipitation_typ
            FROM ({sql_p_daily_winter}) tsp_d_wi
            LEFT JOIN timeseries."{stid}_t" tst
                ON tst.timestamp=tsp_d_wi.date
        """.format(
            sql_p_daily_winter=sql_p_daily_winter,
            **sql_format_dict
        )

        # calculate the delta n
        sql_delta_n = """
            SELECT date,
                CASE WHEN "count_n"> 0 THEN
                        round(("b_{richter_class}" * ("p_d"::float/{n_decim})^"e" * {n_decim})/"count_n")::int
                    ELSE 0
                END AS "delta_10min"
            FROM ({sql_p_daily_precip_class}) tsp_d2
            LEFT JOIN richter_parameters r
                ON r."precipitation_typ"=tsp_d2."precipitation_typ"
        """.format(
            sql_p_daily_precip_class=sql_p_daily_precip_class,
            **sql_format_dict
        )

        # calculate the new corr
        sql_new_corr = """
            SELECT timestamp,
                CASE WHEN "filled" > 0
                     THEN ts."filled" + ts_delta_n."delta_10min"
                     ELSE ts."filled"
                END as corr
            FROM timeseries."{stid}_{para}" ts
            LEFT JOIN ({sql_delta_n}) ts_delta_n
                ON (ts.timestamp)::date = ts_delta_n.date
            {period_clause}
        """.format(
            sql_delta_n=sql_delta_n,
            **sql_format_dict
        )

        # update the timeseries
        sql_update = """
            UPDATE timeseries."{stid}_{para}" ts
            SET "corr" = new.corr
            FROM ({sql_new_corr}) new
            WHERE ts.timestamp = new.timestamp
                AND ts.corr is distinct from new.corr;
        """.format(
            sql_new_corr=sql_new_corr,
            **sql_format_dict
        )

        # run commands
        if "return_sql" in kwargs and kwargs["return_sql"]:
            return sql_update

        self._execute_long_sql(
            sql_update,
            description="richter corrected for the period {min_tstp} - {max_tstp}".format(
                **period.get_sql_format_dict(format=self._tstp_format_human)
            ))

        # mark last import as done, if previous are ok
        if (self.is_last_imp_done(kind="qc") and self.is_last_imp_done(kind="filled")):
            if (period_in.is_empty() or
                period_in.contains(self.get_last_imp_period())):
                self._mark_last_imp_done(kind="corr")

        # update filled time in meta table
        self.update_period_meta(kind="corr")

        # update multi annual mean
        self.update_ma_timeseries(kind="corr")

    @db_engine.deco_update_privilege
    def corr(self, *args, **kwargs):
        return self.richter_correct(*args, **kwargs)

    @db_engine.deco_update_privilege
    def last_imp_richter_correct(self, _last_imp_period=None, **kwargs):
        """Do the richter correction of the last import.

        Parameters
        ----------
        _last_imp_period : weatherDB.utils.TimestampPeriod, optional
            Give the overall period of the last import.
            This is only for intern use of the stationsN method to not compute over and over again the period.
            The default is None.
        **kwargs : dict, optional
            Additional keyword arguments passed to the richter_correct method.
        """
        if not self.is_last_imp_done(kind="corr"):
            if _last_imp_period is None:
                period = self.get_last_imp_period(all=True)
            else:
                period = _last_imp_period

            self.richter_correct(
                period=period,
                **kwargs
            )

        else:
            log.info("The last import of {para_long} Station {stid} was already richter corrected and is therefor skiped".format(
                stid=self.id, para_long=self._para_long
            ))

    @db_engine.deco_update_privilege
    def last_imp_corr(self, _last_imp_period=None, **kwargs):
        """A wrapper for last_imp_richter_correct()."""
        return self.last_imp_richter_correct(_last_imp_period=_last_imp_period, **kwargs)

    @db_engine.deco_update_privilege
    def _sql_fillup_extra_dict(self, **kwargs):
        fillup_extra_dict = super()._sql_fillup_extra_dict(**kwargs)

        stat_pd = StationPD(self.id)
        if stat_pd.isin_db() and \
                not stat_pd.get_filled_period(kind="filled", from_meta=True).is_empty():
            # adjust 10 minutes sum to match measured daily value,
            # but don't add more than 10mm/10min and don't create single peaks with more than 5mm/min
            sql_extra = """
                UPDATE new_filled_{stid}_{para} ts
                SET filled = tsnew.filled
                FROM (
                    SELECT ts.timestamp,
                        CASE WHEN tsb.filled = 0 AND tsa.filled = 0
                             THEN LEAST(ts.filled * coef, 5*{decim})
                             ELSE CASE WHEN ((ts.filled * coef) - ts.filled) <= (10 * {decim})
                                       THEN LEAST(ts.filled * coef, 50*{decim})
                                       ELSE LEAST(ts.filled + (10 * {decim}), 50*{decim})
                                  END
                             END as filled
                    FROM new_filled_{stid}_{para} ts
                    INNER JOIN (
                        SELECT
                            date,
                            ts_d."raw"/ts_10."filled"::float AS coef
                        FROM (
                            SELECT
                                date(timestamp - '5h 50min'::INTERVAL),
                                sum(filled) AS filled
                            FROM new_filled_{stid}_{para}
                            GROUP BY date(timestamp - '5h 50min'::INTERVAL)
                            ) ts_10
                        LEFT JOIN timeseries."{stid}_p_d" ts_d
                            ON ts_10.date=ts_d.timestamp
                        WHERE ts_d."raw" IS NOT NULL
                            AND ts_10.filled > 0
                        ) df_coef
                        ON (ts.timestamp - '5h 50min'::INTERVAL)::date = df_coef.date
                           AND coef != 1
                    LEFT JOIN timeseries."{stid}_{para}" tsb
                        ON ts.timestamp = tsb.timestamp - INTERVAL '10 min'
                    LEFT JOIN timeseries."{stid}_{para}" tsa
                        ON ts.timestamp = tsa.timestamp + INTERVAL '10 min'
                    WHERE ts.filled IS NOT NULL
                ) tsnew
                WHERE tsnew.timestamp = ts.timestamp;
            """.format(stid=self.id, para=self._para, decim=self._decimals)

            fillup_extra_dict.update(dict(sql_extra_after_loop=sql_extra))
        else:
            log.warning(f"StationP({self.id}).fillup: There is no daily timeserie in the database, "+
                        "therefor the 10 minutes values are not getting adjusted to daily values")

        return fillup_extra_dict

    def get_corr(self, **kwargs):
        return self.get_df(kinds=["corr"], **kwargs)

    def get_qn(self, **kwargs):
        return self.get_df(kinds=["qn"], **kwargs)

    def get_richter_class(self, update_if_fails=True):
        """Get the richter class for this station.

        Provide the data from the meta table.

        Parameters
        ----------
        update_if_fails: bool, optional
            Should the richter class get updatet if no exposition class is found in the meta table?
            If False and no exposition class was found None is returned.
            The default is True.

        Returns
        -------
        string
            The corresponding richter exposition class.
        """
        sql = """
            SELECT richter_class
            FROM meta_{para}
            WHERE station_id = {stid}
        """.format(stid=self.id, para=self._para)

        with db_engine.connect() as con:
            res = con.execute(sqltxt(sql)).first()

        # check result
        if res is None:
            if update_if_fails:
                if db_engine.is_superuser:
                    self.update_richter_class()
                    # update_if_fails is False to not get an endless loop
                    return self.get_richter_class(update_if_fails=False)
                else:
                    warnings.warn("You don't have the permissions to change something on the database.\nTherefor an update of the richter_class is not possible.")
                    return None
            else:
                return None
        else:
            return res[0]

    def get_horizon(self):
        """Get the value for the horizon angle. (Horizontabschirmung)

        This value is defined by Richter (1995) as the mean horizon angle in the west direction as:
        H’=0,15H(S-SW) +0,35H(SW-W) +0,35H(W-NW) +0, 15H(NW-N)

        Returns
        -------
        float or None
            The mean western horizon angle
        """
        return self.get_meta(infos="horizon")




# libraries
import logging
import sqlalchemy as sa
from sqlalchemy import text as sqltxt
from functools import cached_property

from ..db.connections import db_engine
from ..utils.dwd import dwd_id_to_str
from ..db.models import MetaT
from .StationBases import StationTETBase

# set settings
# ############
__all__ = ["StationT"]
log = logging.getLogger(__name__)

# class definition
##################
class StationT(StationTETBase):
    """A class to work with and download temperaure data for one station."""

    # common settings
    _MetaModel = MetaT
    _para = "t"
    _para_base = _para
    _para_long = "Temperature"
    _unit = "°C"
    _decimals = 10
    _valid_kinds = {"raw", "raw_min", "raw_max", "qc",
                    "filled", "filled_min", "filled_max", "filled_by"}

    # cdc dwd parameters
    _ftp_folder_base = [
        "climate_environment/CDC/observations_germany/climate/daily/kl/"]
    _cdc_date_col = "MESS_DATUM"
    _cdc_col_names_imp = ["TMK", "TNK", "TXK"]
    _db_col_names_imp = ["raw", "raw_min", "raw_max"]

    # aggregation
    _agg_fun = "avg"

    # for regionalistaion
    _ma_terms = ["year"]
    _coef_sign = ["-", "+"]

    # # for the fillup
    _filled_by_n = 5
    _fillup_max_dist = 100e3


    def __init__(self, id, **kwargs):
        super().__init__(id, **kwargs)
        self.id_str = dwd_id_to_str(id)

    @cached_property
    def _table(self):
        return sa.table(
            f"{self.id}_{self._para}",
            sa.column("timestamp", sa.Date),
            sa.column("raw", sa.Integer),
            sa.column("raw_min", sa.Integer),
            sa.column("raw_max", sa.Integer),
            sa.column("qc", sa.Integer),
            sa.column("filled", sa.Integer),
            sa.column("filled_min", sa.Integer),
            sa.column("filled_max", sa.Integer),
            sa.column("filled_by", sa.SmallInteger),
            schema="timeseries")

    def _create_timeseries_table(self):
        """Create the timeseries table in the DB if it is not yet existing."""
        sql_add_table = f'''
            CREATE TABLE IF NOT EXISTS timeseries."{self.id}_{self._para}"  (
                timestamp date PRIMARY KEY,
                raw integer NULL DEFAULT NULL,
                raw_min integer NULL DEFAULT NULL,
                raw_max integer NULL DEFAULT NULL,
                qc integer NULL DEFAULT NULL,
                filled integer NULL DEFAULT NULL,
                filled_min integer NULL DEFAULT NULL,
                filled_max integer NULL DEFAULT NULL,
                filled_by smallint[{self._filled_by_n}] NULL DEFAULT NULL
                );
        '''
        with db_engine.connect() as con:
            con.execute(sqltxt(sql_add_table))
            con.commit()

    def _get_sql_new_qc(self, period):
        # inversion possible?
        do_invers = self.get_meta(infos=["stationshoehe"])>800

        sql_nears = self._get_sql_near_median(
            period=period, only_real=False, add_is_winter=do_invers,
            extra_cols="raw-nbs_median AS diff")

        if do_invers:
            # without inversion
            sql_null_case = "CASE WHEN (winter) THEN "+\
                f"diff < {-5 * self._decimals} ELSE "+\
                f"ABS(diff) > {5 * self._decimals} END "+\
                f"OR raw < {-50 * self._decimals} OR raw > {50 * self._decimals}"
        else:
            # with inversion
            sql_null_case = f"ABS(diff) > {5 * self._decimals}"

        # create sql for new qc
        sql_new_qc = f"""
            WITH nears AS ({sql_nears})
            SELECT
                timestamp,
                (CASE WHEN ({sql_null_case})
                    THEN NULL
                    ELSE nears."raw"
                    END) as qc
            FROM nears
        """

        return sql_new_qc

    @db_engine.deco_update_privilege
    def _sql_fillup_extra_dict(self, **kwargs):
        # additional parts to calculate the filling of min and max
        fillup_extra_dict = super()._sql_fillup_extra_dict(**kwargs)
        sql_array_init = "ARRAY[{0}]".format(
            ", ".join(["NULL::smallint"] * self._filled_by_n))
        fillup_extra_dict.update({
            "extra_new_temp_cols": "raw_min AS filled_min, raw_max AS filled_max," +
                        f"{sql_array_init} AS nb_min, {sql_array_init} AS nb_max,",
            "extra_cols_fillup_calc": "filled_min=round(nb.raw_min + %3$s, 0)::int, " +
                                      "filled_max=round(nb.raw_max + %3$s, 0)::int, ",
            "extra_cols_fillup": "filled_min = new.filled_min, " +
                                 "filled_max = new.filled_max, ",
            "extra_fillup_where": ' OR ts."filled_min" IS DISTINCT FROM new."filled_min"' +
                                  ' OR ts."filled_max" IS DISTINCT FROM new."filled_max"',
            "extra_exec_cols": "nb_max[{i}]=round(nb.raw_max + %3$s, 0)::int,"+
                               "nb_min[{i}]=round(nb.raw_min + %3$s, 0)::int,",
            "extra_after_loop_extra_col": """,
                filled_min=(SELECT percentile_cont(0.5) WITHIN GROUP (ORDER BY v)
                            FROM unnest(nb_min) as T(v)),
                filled_max=(SELECT percentile_cont(0.5) WITHIN GROUP (ORDER BY v)
                            FROM unnest(nb_max) as T(v))"""})
        return fillup_extra_dict

    def get_multi_annual_raster(self):
        mas = super().get_multi_annual_raster()
        if mas is not None:
            return [ma / 10 for ma in mas]
        else:
            return None

    def get_adj(self, **kwargs):
        main_df, adj_df, ma, main_df_tr = super().get_adj(**kwargs)

        # calculate the yearly
        main_df_y = main_df.groupby(main_df_tr.index.year)\
            .mean().mean()

        adj_df["adj"] = (main_df + (ma[0] - main_df_y)).round(1)

        return adj_df

    def get_quotient(self, **kwargs):
        raise NotImplementedError("The quotient is not yet implemented for temperature.")


# libraries
import sqlalchemy
from sqlalchemy import text as sqltxt
from sqlalchemy import URL
import functools

from ..config import config

# DB connection
###############
class DBEngine:
    def __init__(self):
        self._reset_engine()

        # add the listeners for configuration changes
        self._config_listener_connection = (
            "database",
            "connection",
            self._connection_update)
        self._config_listener_subsection = (
            f"database:{config.get('database', 'connection')}",
            None,
            self._reset_engine)
        config.add_listener(*self._config_listener_connection)
        config.add_listener(*self._config_listener_subsection)

    def __del__(self):
        config.remove_listener(*self._config_listener_connection)
        config.remove_listener(*self._config_listener_subsection)

    def _connection_update(self):
        """Listener if another database subsection is selected."""
        # first remove the old section listener
        config.remove_listener(*self._config_listener_subsection)

        # add the new section listener
        self._config_listener_subsection = (
            f"database:{config.get('database', 'connection')}",
            None,
            self._reset_engine)
        config.add_listener(*self._config_listener_subsection)

        # create a new engine
        self.create_engine()

    # Engine and Session
    # ##################
    def _reset_engine(self):
        """Reset the engine to None to force a recreation on next usage."""
        self._engine = None
        self._session = None
        self._is_superuser = None
        self._select_privilege = None
        self._create_privilege = None
        self._update_privilege = None
        self._insert_privilege = None
        self._delete_privilege = None
        self._priviledge_to_recheck = set()

    def connect(self, *args, **kwargs):
        return self.get_engine().connect(*args, **kwargs)

    def get_engine(self):
        """Get the sqlalchemy database engine.

        Returns the last created engine if possible or creates a new one.

        Returns
        -------
        sqlalchemy.engine.base.Engine
            _description_
        """
        if self._engine is None:
            self.create_engine()

        return self._engine

    def create_engine(self):
        # create the engine
        con_key = config.get("database", "connection")
        con_sect_key = f"database:{con_key}"
        con_section = config[con_sect_key]
        user, pwd = config.get_db_credentials()
        self._engine = sqlalchemy.create_engine(
            URL.create(
                drivername="postgresql+psycopg2",
                username=user,
                password=pwd,
                host=con_section["HOST"],
                database=con_section["DATABASE"],
                port=con_section["PORT"]),
            connect_args={
                "options": "-c timezone=utc",
                'connect_timeout': 10})
        self._check_is_superuser()

        return self._engine

    @property
    def engine(self):
        return self.get_engine()

    @engine.setter
    def engine(self, value):
        raise PermissionError("You are not allowed to change the engine of the database connection.\nPlease change the configuration to change the database connection.")

    def create_session(self):
        """Create a new session from the engine.

        Returns
        -------
        sqlalchemy.orm.session.Session
            _description_
        """
        self._session = sqlalchemy.orm.sessionmaker(bind=self.get_engine())
        return self._session

    def get_session(self):
        """Get a session from the engine.

        Returns
        -------
        sqlalchemy.orm.session.Session
            _description_
        """
        if self._session is None:
            return self.create_session()
        else:
            return self._session

    @property
    def session(self):
        return self.get_session()

    @session.setter
    def session(self, value):
        raise PermissionError("You are not allowed to change the session of the database connection.\nPlease change the configuration to change the database connection.")

    def reload_config(self):
        """Reload the configuration and create a new engine.
        """
        self.create_engine()

    # Privilege checks
    ##################
    def _check_is_superuser(self):
        with self.connect() as con:
            self._is_superuser = con.execute(sqltxt(
                "SELECT usesuper FROM pg_user WHERE usename = current_user;"
            )).first()[0]
        return self._is_superuser

    @property
    def is_superuser(self):
        if self._is_superuser is None:
            self._check_is_superuser()
        return self._is_superuser

    @is_superuser.setter
    def is_superuser(self, value):
        raise PermissionError("You are not allowed to change the superuser status of the database connection, as this is due to PostGreSQL database privileges of your database user.")

    def _check_privilege(self, privilege):
        if self._is_superuser:
            return True
        else:
            with self.connect() as con:
                if privilege == "CREATE":
                    return con.execute(sqltxt(
                        f"SELECT pg_catalog.has_schema_privilege('public', '{privilege}');"
                    )).first()[0]
                res = con.execute(sqltxt(
                    f"""SELECT bool_and(pg_catalog.has_table_privilege(table_name, '{privilege}'))
                       FROM information_schema.TABLES
                       WHERE table_schema=CURRENT_SCHEMA AND table_type='BASE TABLE';"""
                )).first()[0]
                if res is None:
                    if privilege in self._priviledge_to_recheck:
                        self.log.debug(f"Privilege {privilege} not found, probably because database isn't initiated. Setting as True.")
                    else:
                        self._priviledge_to_recheck.add(privilege)
                    return True
                else:
                    if privilege in self._priviledge_to_recheck:
                        self._priviledge_to_recheck.remove(privilege)
                    return res

    def _check_select_privilege(self):
        """Check on the database if the user has the SELECT privilege."""
        self._select_privilege = self._check_privilege("SELECT")
        return self._select_privilege

    def _check_update_privilege(self):
        """Check on the database if the user has the UPDATE privilege."""
        self._update_privilege = self._check_privilege("UPDATE")
        return self._update_privilege

    def _check_insert_privilege(self):
        """Check on the database if the user has the INSERT privilege."""
        self._insert_privilege = self._check_privilege("INSERT")
        return self._insert_privilege

    def _check_create_privilege(self):
        """Check on the database if the user has the CREATE privilege."""
        self._create_privilege = self._check_privilege("CREATE")
        return self._create_privilege

    def _check_delete_privilege(self):
        """Check on the database if the user has the DELETE privilege."""
        self._delete_privilege = self._check_privilege("DELETE")
        return self._delete_privilege

    @property
    def select_privilege(self):
        """Does the user have the PostGreSQL SELECT privilege on the database?"""
        if self._select_privilege is None or "SELECT" in self._priviledge_to_recheck:
            self._check_select_privilege()
        return self._select_privilege

    @property
    def update_privilege(self):
        """Does the user have the PostGreSQL UPDATE privilege on the database?"""
        if self._update_privilege is None or "UPDATE" in self._priviledge_to_recheck:
            self._check_update_privilege()
        return self._update_privilege

    @property
    def insert_privilege(self):
        """Does the user have the PostGreSQL INSERT privilege on the database?"""
        if self._insert_privilege is None or "INSERT" in self._priviledge_to_recheck:
            self._check_insert_privilege()
        return self._insert_privilege

    @property
    def upsert_privilege(self):
        """Does the user have the PostGreSQL INSERT and UPDATE privilege on the database?"""
        return self.insert_privilege and self.update_privilege

    @property
    def create_privilege(self):
        """Does the user have the PostGreSQL CREATE privilege on the database?"""
        if self._create_privilege is None or "CREATE" in self._priviledge_to_recheck:
            self._check_create_privilege()
        return self._create_privilege

    @property
    def delete_privilege(self):
        """Does the user have the PostGreSQL DELETE privilege on the database?"""
        if self._delete_privilege is None or "DELETE" in self._priviledge_to_recheck:
            self._check_delete_privilege()
        return self._delete_privilege

    @property
    def all_privileges(self):
        """Does the user have all (SELECT, UPDATE, INSERT, DELETE, CREATE) PostGreSQL privileges on the database?"""
        return self.select_privilege and \
            self.update_privilege and \
            self.insert_privilege and \
            self.create_privilege and \
            self.delete_privilege

    def _privilege_setters(self, property_name):
        raise PermissionError(f"You are not allowed to change the value of {property_name} as this is due to PostGreSQL database privileges of your user.")

    @select_privilege.setter
    def select_privilege(self, value):
        self._privilege_setters("select_privilege")

    @update_privilege.setter
    def update_privilege(self, value):
        self._privilege_setters("update_privilege")

    @insert_privilege.setter
    def insert_privilege(self, value):
        self._privilege_setters("insert_privilege")

    @upsert_privilege.setter
    def upsert_privilege(self, value):
        self._privilege_setters("upsert_privilege")

    @create_privilege.setter
    def create_privilege(self, value):
        self._privilege_setters("create_privilege")

    @delete_privilege.setter
    def delete_privilege(self, value):
        self._privilege_setters("delete_privilege")

    @all_privileges.setter
    def all_privileges(self, value):
        self._privilege_setters("all_privileges")

    def deco_is_superuser(self, target):
        """Decorator to check if the user is a superuser."""
        @functools.wraps(target)
        def wrapper(*args, **kwargs):
            if self.is_superuser:
                return target(*args, **kwargs)
            else:
                raise PermissionError("You are no super user of the Database and therefor this function is not available.")
        return wrapper

    def deco_select_privilege(self, target):
        """Decorator to check if the user has the SELECT privilege."""
        @functools.wraps(target)
        def wrapper(*args, **kwargs):
            if self.select_privilege:
                return target(*args, **kwargs)
            else:
                raise PermissionError("You have no select privilege on the database and therefor this function is not available.")
        return wrapper

    def deco_update_privilege(self, target):
        """Decorator to check if the user has the UPDATE privilege."""
        @functools.wraps(target)
        def wrapper(*args, **kwargs):
            if self.update_privilege:
                return target(*args, **kwargs)
            else:
                raise PermissionError("You have no update privilege on the database and therefor this function is not available.")
        return wrapper

    def deco_insert_privilege(self, target):
        """Decorator to check if the user has the INSERT privilege."""
        @functools.wraps(target)
        def wrapper(*args, **kwargs):
            if self.insert_privilege:
                return target(*args, **kwargs)
            else:
                raise PermissionError("You have no insert privilege on the database and therefor this function is not available.")
        return wrapper

    def deco_upsert_privilege(self, target):
        """Decorator to check if the user has the INSERT and UPDATE privilege."""
        @functools.wraps(target)
        def wrapper(*args, **kwargs):
            if self.upsert_privilege:
                return target(*args, **kwargs)
            else:
                raise PermissionError("You have no upsert privilege on the database and therefor this function is not available.")
        return wrapper

    def deco_create_privilege(self, target):
        """Decorator to check if the user has the CREATE privilege."""
        @functools.wraps(target)
        def wrapper(*args, **kwargs):
            if self.create_privilege:
                return target(*args, **kwargs)
            else:
                raise PermissionError("You are no admin user of the Database and therefor this function is not available.")
        return wrapper

    def deco_delete_privilege(self, target):
        """Decorator to check if the user has the DELETE privilege."""
        @functools.wraps(target)
        def wrapper(*args, **kwargs):
            if self.delete_privilege:
                return target(*args, **kwargs)
            else:
                raise PermissionError("You have no delete privilege on the database and therefor this function is not available.")
        return wrapper

    def deco_all_privileges(self, target):
        """Decorator to check if the user has all (SELECT, UPDATE, INSERT, DELETE, CREATE) privileges."""
        @functools.wraps(target)
        def wrapper(*args, **kwargs):
            if self.all_privileges:
                return target(*args, **kwargs)
            else:
                raise PermissionError("You are no super user of the Database and therefor this function is not available.")
        return wrapper


db_engine = DBEngine()



# libraries
import logging
import sqlalchemy as sa
from sqlalchemy import text as sqltxt
from functools import cached_property

from ..db.connections import db_engine
from ..db.models import MetaET
from .StationBases import StationTETBase

# set settings
# ############
__all__ = ["StationET"]
log = logging.getLogger(__name__)

# class definition
##################
class StationET(StationTETBase):
    """A class to work with and download potential Evapotranspiration (VPGB) data for one station."""

    # common settings
    _MetaModel = MetaET
    _para = "et"
    _para_base = _para
    _para_long = "potential Evapotranspiration"
    _unit = "mm/Tag"
    _decimals = 10

    # cdc dwd parameters
    _ftp_folder_base = ["climate_environment/CDC/derived_germany/soil/daily/"]
    _ftp_zip_regex_prefix = r".*_v2_"
    _cdc_date_col = "Datum"
    _cdc_col_names_imp = ["VPGFAO"]

    # for regionalistaion
    _ma_terms = ["year"]

    # for the fillup
    _fillup_max_dist = 100000

    def __init__(self, id, **kwargs):
        super().__init__(id, **kwargs)
        self.id_str = str(id)

    @cached_property
    def _table(self):
        return sa.table(
            f"{self.id}_{self._para}",
            sa.column("timestamp", sa.Date),
            sa.column("raw", sa.Integer),
            sa.column("qc", sa.Integer),
            sa.column("filled", sa.Integer),
            sa.column("filled_by", sa.SmallInteger),
            schema="timeseries")

    def _create_timeseries_table(self):
        """Create the timeseries table in the DB if it is not yet existing."""
        sql_add_table = '''
            CREATE TABLE IF NOT EXISTS timeseries."{stid}_{para}"  (
                timestamp date PRIMARY KEY,
                raw integer NULL DEFAULT NULL,
                qc integer NULL DEFAULT NULL,
                filled integer NULL DEFAULT NULL,
                filled_by smallint NULL DEFAULT NULL
                );
        '''.format(stid=self.id, para=self._para)
        with db_engine.connect() as con:
            con.execute(sqltxt(sql_add_table))
            con.commit()

    def _get_sql_new_qc(self, period):
        # inversion possible?
        do_invers = self.get_meta(infos=["stationshoehe"])>800

        sql_nears = self._get_sql_near_median(
            period=period, only_real=False, add_is_winter=do_invers,
            extra_cols="raw-nbs_median AS diff")

        sql_null_case = f"""(nears.raw > (nears.nbs_median * 2) AND nears.raw > {3*self._decimals})
                            OR ((nears.raw * 4) < nears.nbs_median AND nears.raw > {2*self._decimals})"""
        if do_invers:
            # without inversion
            sql_null_case = "CASE WHEN (winter) THEN "+\
                f"((nears.raw * 4) < nears.nbs_median AND nears.raw > {2*self._decimals}) ELSE "+\
                f"{sql_null_case} END"

        # create sql for new qc
        sql_new_qc = f"""
            WITH nears AS ({sql_nears})
            SELECT
                timestamp,
                (CASE WHEN ({sql_null_case}
                            OR (nears.raw < 0)
                            OR (nears.raw > {20*self._decimals}))
                    THEN NULL
                    ELSE nears."raw" END) as qc
            FROM nears
        """

        return sql_new_qc

    def get_adj(self, **kwargs):
        main_df, adj_df, ma, main_df_tr = super().get_adj(**kwargs)

        # calculate the yearly
        main_df_y = main_df.groupby(main_df_tr.index.year)\
            .sum(min_count=345).mean()

        adj_df["adj"] = (main_df * (ma[0] / main_df_y)).round(1)

        return adj_df



# libraries
from datetime import datetime, timezone

# Variables
# ---------
# The minimal timestamp that is allowed for the data
MIN_TSTP = datetime.strptime("19940101", "%Y%m%d").replace(tzinfo=timezone.utc)

# possible aggregation periods from small to big
AGG_TO = {
    None: {
        "split":{"n": 5, "t":3, "et": 3}},
    "10 min": {
        "split":{"n": 5, "t":3, "et": 3}},
    "hour": {
        "split":{"n": 4, "t":3, "et": 3}},
    "day": {
        "split":{"n": 3, "t":3, "et": 3}},
    "month": {
        "split":{"n": 2, "t":2, "et": 2}},
    "year": {
        "split":{"n": 1, "t":1, "et": 1}},
    "decade": {
        "split":{"n": 1, "t":1, "et": 1}}
    }

__all__ = ["MIN_TSTP", "AGG_TO"]


__author__ = "Max Schmit"
__email__ = "max.schmit@hydrology.uni-freiburg.de"
__copyright__ = "Copyright 2024, Max Schmit"

try:
    from ._version import __version__
except ImportError:
    __version__ = "0.0.0" # not set when running from source

from .utils.logging import remove_old_logs, setup_logging_handlers
from . import station, stations, broker
from .station import StationP, StationPD, StationT, StationET, GroupStation
from .stations import StationsP, StationsPD, StationsT, StationsET, GroupStations
from .config import config
from .broker import Broker


remove_old_logs()
setup_logging_handlers()

__all__ = ["StationP", "StationPD", "StationT", "StationET", "GroupStation",
           "StationsP", "StationsPD", "StationsT", "StationsET", "GroupStations",
           "Broker",
           "station", "stations", "broker", "config"]


import logging
from logging.handlers import TimedRotatingFileHandler
import datetime
from pathlib import Path
import re
import socket
import os

from ..config import config

try:
    import coloredlogs
    cl_available = True
except ImportError:
    cl_available = False

# set the log
#############
log = logging.getLogger(__name__.split(".")[0])

def _get_log_dir ():
    return Path(config.get("logging", "directory"))

def remove_old_logs(max_days=14):
    # remove old logs
    log_dir = _get_log_dir()
    log_date_min = datetime.datetime.now() - datetime.timedelta(days=max_days)
    for log_file in [
            file for file in log_dir.glob("*.log.*")
                if re.match(r".*\.log\.\d{4}-\d{2}-\d{2}$", file.name)]:
        try:
            file_date = datetime.datetime.strptime(log_file.name.split(".")[-1], "%Y-%m-%d")
            if file_date < log_date_min:
                log_file.unlink()
        except:
            pass

def setup_logging_handlers():
    """Setup the logging handlers depending on the configuration.

    Raises
    ------
    ValueError
        If the handler type is not known.
    """
    # get log dir
    log_dir = _get_log_dir()
    if not log_dir.is_dir(): log_dir.mkdir()

    # add filehandler if necessary
    log.setLevel(config.get("logging", "level", fallback=logging.DEBUG))
    handler_names = [h.get_name() for h in log.handlers]
    format = config.get(
        "logging",
        "format",
        raw=True,
        fallback="%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    level = config.get("logging", "level", fallback=logging.DEBUG)
    for handler_type in config.getlist("logging", "handlers"):
        handler_name = f"weatherDB_config:{handler_type}"

        # check if coloredlogs is available
        if cl_available and handler_type == "console":
            coloredlogs.install(level=level, fmt=format, logger=log)
            log.debug("Using coloredlogs")
            continue

        # get log file name
        if handler_type == "file":
            try:
                user = os.getlogin()
            except:
                user = "anonym"
            host = socket.gethostname().replace(".","_")
            log_file = log_dir.joinpath(
                config.get("logging", "file", fallback="weatherDB_{user}_{host}.log")\
                    .format(user=user, host=host))

        # get or create handler
        if handler_name not in handler_names:
            if handler_type == "console":
                handler = logging.StreamHandler()
            elif handler_type == "file":
                handler = TimedRotatingFileHandler(
                    log_file,
                    when="midnight",
                    encoding="utf-8")
            else:
                raise ValueError(f"Handler '{handler_type}' not known.")

            handler.set_name(handler_name)
            log.addHandler(handler)

        elif handler_name in handler_names:
            handler = log.handlers[handler_names.index(handler_name)]

            # check if file path has changed
            if handler_type == "file" and handler.baseFilename != str(log_file):
                log.removeHandler(handler)
                handler.close()
                handler = TimedRotatingFileHandler(
                    log_file,
                    when="midnight",
                    encoding="utf-8")
                handler.set_name(handler_name)
                log.addHandler(handler)

        # set formatter and level
        handler.setFormatter(
            logging.Formatter(format))
        handler.setLevel(level)

# add config listener
config.add_listener("logging", None, setup_logging_handlers)


"""
Some utilities functions to download the needed data for the module to work.
"""
import requests
from pathlib import Path
from distutils.util import strtobool
import hashlib
import progressbar as pb

from ..config import config

def download_ma_rasters(which="all", overwrite=None, update_user_config=False):
    """Get the multi annual rasters on which bases the regionalisation is done.

    The refined multi annual datasets, that are downloaded are published on Zenodo [1]_

    References
    ----------
    .. [1] Schmit, M.; Weiler, M. (2023). German weather services (DWD) multi annual meteorological rasters for the climate period 1991-2020 refined to 25m grid (1.0.0) [Data set]. Zenodo. https://doi.org/10.5281/zenodo.10066045

    Parameters
    ----------
    which : str or [str], optional
        Which raster to download.
        Options are "dwd", "hyras", "regnie" and "all".
        The default is "all".
    overwrite : bool, optional
        Should the multi annual rasters be downloaded even if they already exist?
        If None the user will be asked.
        The default is None.
    update_user_config : bool, optional
        Should the downloaded rasters be set as the regionalisation rasters in the user configuration file?
        The default is False.
    """
    # DOI of the multi annual dataset
    DOI = "10.5281/zenodo.10066045"

    # check which
    if isinstance(which, str):
        which = [which]
    for w in which:
        if w not in ["all", "dwd", "hyras", "regnie"]:
            raise ValueError(
                "which must be one of 'all', 'dwd', 'hyras' or 'regnie'.")
        if w == "all":
            which = ["dwd", "hyras", "regnie"]
            break

    # get zenodo record
    zenodo_id = requests.get(
        f"https://doi.org/{DOI}"
    ).url.split("/")[-1]
    zenodo_rec = requests.get(
        f"https://zenodo.org/api/records/{zenodo_id}"
    ).json()

    # download files
    for file in zenodo_rec["files"]:
        file_key = file["key"].lower().split("_")[0].split("-")[0]
        if file_key in which:
            # check if file is in config
            if f"data:rasters:{file_key}" not in config:
                print(f"Skipping {file_key} as it is not in your configuration.\nPlease add a section 'data:rasters:{file_key}' to your configuration file.")
                continue

            # check if file already exists
            file_path = Path(config.get(f"data:rasters:{file_key}", "file"))
            if file_path.exists():
                skip = False
                if overwrite is False:
                    skip = True
                elif overwrite is None:
                    skip = not strtobool(input(
                        f"{file_key} already exists at {file_path}.\n"+
                        "Do you want to overwrite it? [y/n] "))

                if skip:
                    print(f"Skipping {file_key} as overwriting is not allowed.")
                    continue

            # check if the directory exists
            if not file_path.parent.exists():
                if strtobool(input(
                        f"The directory \"{file_path.parent}\" does not exist.\n"+
                        "Do you want to create it? [y/n] ")):
                    file_path.parent.mkdir(parents=True)

            # download file
            r = requests.get(file["links"]["self"], stream=True)
            if r.status_code != 200:
                r.raise_for_status()  # Will only raise for 4xx codes, so...
                raise RuntimeError(
                    f'Request to {file["links"]["self"]} returned status code {r.status_code}')
            block_size = 1024
            file_size = int(r.headers.get('Content-Length', 0))
            pbar = pb.ProgressBar(
                max_value=file_size,
                prefix=f"downloading {file_key}: ",
                widgets=[ " ",
                    pb.widgets.DataSize(),
                    "/",
                    pb.widgets.DataSize("max_value"),
                    pb.widgets.AdaptiveTransferSpeed(
                        format='(%(scaled)5.1f %(prefix)s%(unit)-s/s) '),
                    pb.widgets.Bar(), " ",
                    pb.widgets.Percentage(),
                    pb.widgets.ETA()],
                line_breaks=False,
                redirect_stdout=True
                ).start()
            md5 = hashlib.md5()
            with open(file_path, "wb+") as f:
                for i, chunk in enumerate(r.iter_content(block_size)):
                    f.write(chunk)
                    md5.update(chunk)
                    pbar.update(i*block_size)
            pbar.finish()

            # check checksum
            if md5.hexdigest() != file["checksum"].replace("md5:", ""):
                raise ValueError(
                    f"Checksum of {file_key} doesn't match. File might be corrupted.")

            # update user config
            if update_user_config:
                if config.has_user_config:
                    config.update_user_config(f"data:rasters:{file_key}", "file", str(file_path))
                else:
                    print(f"No user configuration file found, therefor the raster '{file_key}' is not set in the user configuration file.")


def download_dem(overwrite=None, extent=(5.3, 46.1, 15.6, 55.4), update_user_config=False):
    """Download the newest DEM data from the Copernicus Sentinel dataset.

    Only the GLO-30 DEM, which has a 30m resolution, is downloaded as it is freely available.
    If you register as a scientific researcher also the EEA-10, with 10 m resolution, is available.
    You will have to download the data yourself and define it in the configuration file.

    After downloading the data, the files are merged and saved as a single tif file in the data directory in a subfolder called 'DEM'.
    To use the DEM data in the WeatherDB, you will have to define the path to the tif file in the configuration file.

    Source:
    Copernicus DEM - Global and European Digital Elevation Model. Digital Surface Model (DSM) provided in 3 different resolutions (90m, 30m, 10m) with varying geographical extent (EEA: European and GLO: global) and varying format (INSPIRE, DGED, DTED). DOI:10.5270/ESA-c5d3d65.

    Parameters
    ----------
    overwrite : bool, optional
        Should the DEM data be downloaded even if it already exists?
        If None the user will be asked.
        The default is None.
    extent : tuple, optional
        The extent in WGS84 of the DEM data to download.
        The default is the boundary of germany + ~40km = (5.3, 46.1, 15.6, 55.4).
    update_user_config : bool, optional
        Should the downloaded DEM be set as the used DEM in the user configuration file?
        The default is False.
    """
    # import necessary modules
    import rasterio as rio
    from rasterio.merge import merge
    import tarfile
    import shutil
    from tempfile import TemporaryDirectory
    import re
    import json

    # get dem_dir
    base_dir = Path(config.get("data", "base_dir"))
    dem_dir = base_dir / "DEM"
    dem_dir.mkdir(parents=True, exist_ok=True)

    # get available datasets
    prism_url = "https://prism-dem-open.copernicus.eu/pd-desk-open-access/publicDemURLs"
    avl_ds_req = json.loads(
        requests.get(
            prism_url,
            headers={"Accept": "json"}
            ).text
    )
    avl_ds = [{
        "id": e["datasetId"],
        "year": int(e["datasetId"].split("/")[1].split("_")[0]),
        "year_part": int(e["datasetId"].split("/")[1].split("_")[1]),
        "resolution": int(e["datasetId"].split("-")[2]),
        } for e in avl_ds_req]

    # select newest and highest resolution dataset
    ds_id = sorted(
        avl_ds,
        key=lambda x: (-x["resolution"], x["year"], x["year_part"])
        )[-1]["id"]

    # check if dataset already exists
    dem_file = dem_dir / f'{ds_id.replace("/", "__")}.tif'
    if dem_file.exists():
        print(f"The DEM data already exists at {dem_file}.")
        if overwrite is None:
            overwrite = strtobool(input("Do you want to overwrite it? [y/n] "))
        if not overwrite:
            print("Skipping, because overwritting was turned of.")
            return
        else:
            print("Overwriting the dataset.")
    dem_dir.mkdir(exist_ok=True)

    # selecting DEM tiles
    print(f"getting available tiles for Copernicus dataset '{ds_id}'")
    ds_files_req = json.loads(
        requests.get(
            f"{prism_url}/{ds_id.replace('/', '__')}",
            headers={"Accept": "json"}
            ).text
    )
    re_comp = re.compile(r".*/Copernicus_DSM_\d{2}_N\d*_\d{2}_E\d*.*")
    ds_files_all = [
        {"lat": int(Path(f["nativeDemUrl"]).stem.split("_")[3][1:]),
         "long": int(Path(f["nativeDemUrl"]).stem.split("_")[5][1:]),
         **f} for f in ds_files_req if re_comp.match(f["nativeDemUrl"])]
    res_deg = 1
    ds_files = list(filter(
        lambda x: (
            (extent[0] - res_deg) < x["long"] < extent[2] and
            (extent[1] - res_deg) < x["lat"] < extent[3]
            ),
        ds_files_all))

    # download DEM tiles
    print("downloading tiles")
    with TemporaryDirectory() as tmp_dir:
        tmp_dir_fp = Path(tmp_dir)
        for f in pb.progressbar(ds_files):
            with open(tmp_dir_fp / Path(f["nativeDemUrl"]).name, "wb") as d:
                d.write(requests.get(f["nativeDemUrl"]).content)
        print("downloaded all files")

        # extracting tifs from tars
        for i, f in pb.progressbar(list(enumerate(tmp_dir_fp.glob("*.tar")))):
            with tarfile.open(f) as t:
                # extract dem tif
                re_comp = re.compile(r"^.*\/DEM\/.*\.tif$")
                name = list(filter(re_comp.match, t.getnames()))[0]
                with open(tmp_dir_fp/f"{name.split('/')[-1]}", "wb") as d:
                    d.write(t.extractfile(name).read())

                # extract info contract
                if i==0:
                    re_comp = re.compile(r"^.*\/INFO\/.*\.pdf$")
                    name = list(filter(re_comp.match, t.getnames()))[0]
                    with open(tmp_dir_fp/f"{name.split('/')[-1]}", "wb") as d:
                        d.write(t.extractfile(name).read())

            # remove tar
            f.unlink()

        # merge files
        srcs = [rio.open(f) for f in tmp_dir_fp.glob("*.tif")]
        dem_np, dem_tr = merge(srcs)
        dem_meta = srcs[0].meta.copy()
        dem_meta.update({
            "driver": "GTiff",
            "height": dem_np.shape[1],
            "width": dem_np.shape[2],
            "transform": dem_tr
        })
        with rio.open(dem_file, "w", **dem_meta) as d:
            d.write(dem_np)

        # copy info contract
        tmp_eula_fp = next(tmp_dir_fp.glob("*.pdf"))
        shutil.copyfile(
            tmp_eula_fp,
            dem_dir / tmp_eula_fp.name
        )

    print(f"created DEM at '{dem_file}'.")

    # update user config
    if update_user_config:
        if config.has_user_config:
            config.update_user_config("data:rasters", "dems", str(dem_file))
            return
        else:
            print("No user configuration file found, therefor the DEM is not set in the user configuration file.")

    print("To use the DEM data in the WeatherDB, you will have to define the path to the tif file in the user configuration file.")


"""
This module has a class for every type of station. E.g. StationP (or StationP).
One object represents one Station with one parameter.
This object can get used to get the corresponding timeserie.
There is also a StationGroup class that groups the three parameters precipitation, temperature and evapotranspiration together for one station.
"""
from .StationET import StationET
from .StationP import StationP
from .StationPD import StationPD
from .StationT import StationT

__all__ = ["StationET", "StationP", "StationPD", "StationT"]


"""Initial database creation

Revision ID: V1.0.0
Revises:
Create Date: 2024-08-23 14:44:43.130167

"""

from typing import Sequence, Union
from alembic import op
import sqlalchemy as sa
import geoalchemy2

# revision identifiers, used by Alembic.
revision: str = "V1.0.0"
down_revision: Union[str, None] = None
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    op.create_table(
        "droped_stations",
        sa.Column(
            "station_id",
            sa.Integer(),
            nullable=False,
            comment="The station id that got droped",
        ),
        sa.Column(
            "parameter",
            sa.VARCHAR(length=3),
            nullable=False,
            comment="The parameter (n,t,et,p_d) of the station that got droped",
        ),
        sa.Column(
            "why",
            sa.Text(),
            nullable=False,
            comment="The reason why the station got droped",
        ),
        sa.Column(
            "timestamp",
            sa.TIMESTAMP(),
            server_default=sa.text("now()"),
            nullable=False,
            comment="The timestamp when the station got droped",
        ),
        sa.PrimaryKeyConstraint("station_id", "parameter"),
        comment="This table is there to save the station ids that got droped, so they wont GET recreated",
    )
    op.create_table(
        "meta_et",
        sa.Column(
            "station_id",
            sa.Integer(),
            nullable=False,
            comment="official DWD-ID of the station",
        ),
        sa.Column(
            "is_real",
            sa.Boolean(),
            nullable=False,
            server_default="true",
            comment=" 'Is this station a real station with own measurements or only a virtual station, to have complete timeseries for every precipitation station.",
        ),
        sa.Column(
            "raw_from",
            sa.TIMESTAMP(),
            nullable=True,
            comment='The timestamp from when on own "raw" data is available',
        ),
        sa.Column(
            "raw_until",
            sa.TIMESTAMP(),
            nullable=True,
            comment='The timestamp until when own "raw" data is available',
        ),
        sa.Column(
            "hist_until",
            sa.TIMESTAMP(),
            nullable=True,
            comment='The timestamp until when own "raw" data is available',
        ),
        sa.Column(
            "filled_from",
            sa.TIMESTAMP(),
            nullable=True,
            comment="The timestamp from when on filled data is available",
        ),
        sa.Column(
            "filled_until",
            sa.TIMESTAMP(),
            nullable=True,
            comment="The timestamp until when filled data is available",
        ),
        sa.Column(
            "last_imp_from",
            sa.TIMESTAMP(),
            nullable=True,
            comment="The minimal timestamp of the last import, that might not yet have been treated",
        ),
        sa.Column(
            "last_imp_until",
            sa.TIMESTAMP(),
            nullable=True,
            comment="The maximal timestamp of the last import, that might not yet have been treated",
        ),
        sa.Column(
            "last_imp_filled",
            sa.Boolean(),
            nullable=False,
            server_default="false",
            comment="Got the last import already filled?",
        ),
        sa.Column(
            "stationshoehe",
            sa.Integer(),
            nullable=False,
            comment="The stations height above the ground in meters",
        ),
        sa.Column(
            "stationsname",
            sa.VARCHAR(length=50),
            nullable=False,
            comment="The stations official name as text",
        ),
        sa.Column(
            "bundesland",
            sa.VARCHAR(length=30),
            nullable=False,
            comment="The state the station is located in",
        ),
        sa.Column(
            "geometry",
            geoalchemy2.types.Geometry(
                geometry_type="POINT",
                srid=4326,
                from_text="ST_GeomFromEWKT",
                name="geometry",
                nullable=False,
            ),
            nullable=False,
            comment="The stations location in the WGS84 coordinate reference system (EPSG:4326)",
        ),
        sa.Column(
            "geometry_utm",
            geoalchemy2.types.Geometry(
                geometry_type="POINT",
                srid=25832,
                from_text="ST_GeomFromEWKT",
                name="geometry",
                nullable=False,
            ),
            nullable=False,
            comment="The stations location in the UTM32 coordinate reference system (EPSG:25832)",
        ),
        sa.Column(
            "last_imp_qc",
            sa.Boolean(),
            nullable=False,
            server_default="false",
            comment="Got the last import already quality checked?",
        ),
        sa.Column(
            "qc_from",
            sa.TIMESTAMP(),
            nullable=True,
            comment='The timestamp from when on quality checked("qc") data is available',
        ),
        sa.Column(
            "qc_until",
            sa.TIMESTAMP(),
            nullable=True,
            comment='The timestamp until when quality checked("qc") data is available',
        ),
        sa.Column(
            "qc_droped",
            sa.FLOAT(),
            nullable=True,
            comment="The percentage of droped values during the quality check",
        ),
        sa.PrimaryKeyConstraint("station_id"),
        comment="The Meta informations of the evapotranspiration stations.",
    )
    op.create_index(
        "idx_meta_et_geometry",
        "meta_et",
        ["geometry"],
        unique=False,
        postgresql_using="gist",
    )
    op.create_index(
        "idx_meta_et_geometry_utm",
        "meta_et",
        ["geometry_utm"],
        unique=False,
        postgresql_using="gist",
    )
    op.create_table(
        "meta_p",
        sa.Column(
            "last_imp_corr",
            sa.Boolean(),
            nullable=False,
            server_default="false",
            comment="Got the last import already Richter corrected?",
        ),
        sa.Column(
            "corr_from",
            sa.TIMESTAMP(),
            nullable=True,
            comment="The timestamp from when on corrected data is available",
        ),
        sa.Column(
            "corr_until",
            sa.TIMESTAMP(),
            nullable=True,
            comment="The timestamp until when corrected data is available",
        ),
        sa.Column(
            "horizon",
            sa.FLOAT(),
            nullable=True,
            comment="The horizon angle in degrees, how it got defined by Richter(1995).",
        ),
        sa.Column(
            "richter_class",
            sa.String(),
            nullable=True,
            comment="The Richter exposition class, that got derived from the horizon angle.",
        ),
        sa.Column(
            "station_id",
            sa.Integer(),
            nullable=False,
            comment="official DWD-ID of the station",
        ),
        sa.Column(
            "is_real",
            sa.Boolean(),
            nullable=False,
            comment=" 'Is this station a real station with own measurements or only a virtual station, to have complete timeseries for every precipitation station.",
        ),
        sa.Column(
            "raw_from",
            sa.TIMESTAMP(),
            nullable=True,
            comment='The timestamp from when on own "raw" data is available',
        ),
        sa.Column(
            "raw_until",
            sa.TIMESTAMP(),
            nullable=True,
            comment='The timestamp until when own "raw" data is available',
        ),
        sa.Column(
            "hist_until",
            sa.TIMESTAMP(),
            nullable=True,
            comment='The timestamp until when own "raw" data is available',
        ),
        sa.Column(
            "filled_from",
            sa.TIMESTAMP(),
            nullable=True,
            comment="The timestamp from when on filled data is available",
        ),
        sa.Column(
            "filled_until",
            sa.TIMESTAMP(),
            nullable=True,
            comment="The timestamp until when filled data is available",
        ),
        sa.Column(
            "last_imp_from",
            sa.TIMESTAMP(),
            nullable=True,
            comment="The minimal timestamp of the last import, that might not yet have been treated",
        ),
        sa.Column(
            "last_imp_until",
            sa.TIMESTAMP(),
            nullable=True,
            comment="The maximal timestamp of the last import, that might not yet have been treated",
        ),
        sa.Column(
            "last_imp_filled",
            sa.Boolean(),
            nullable=False,
            server_default="false",
            comment="Got the last import already filled?",
        ),
        sa.Column(
            "stationshoehe",
            sa.Integer(),
            nullable=False,
            comment="The stations height above the ground in meters",
        ),
        sa.Column(
            "stationsname",
            sa.VARCHAR(length=50),
            nullable=False,
            comment="The stations official name as text",
        ),
        sa.Column(
            "bundesland",
            sa.VARCHAR(length=30),
            nullable=False,
            comment="The state the station is located in",
        ),
        sa.Column(
            "geometry",
            geoalchemy2.types.Geometry(
                geometry_type="POINT",
                srid=4326,
                from_text="ST_GeomFromEWKT",
                name="geometry",
                nullable=False,
            ),
            nullable=False,
            comment="The stations location in the WGS84 coordinate reference system (EPSG:4326)",
        ),
        sa.Column(
            "geometry_utm",
            geoalchemy2.types.Geometry(
                geometry_type="POINT",
                srid=25832,
                from_text="ST_GeomFromEWKT",
                name="geometry",
                nullable=False,
            ),
            nullable=False,
            comment="The stations location in the UTM32 coordinate reference system (EPSG:25832)",
        ),
        sa.Column(
            "last_imp_qc",
            sa.Boolean(),
            nullable=False,
            server_default="false",
            comment="Got the last import already quality checked?",
        ),
        sa.Column(
            "qc_from",
            sa.TIMESTAMP(),
            nullable=True,
            comment='The timestamp from when on quality checked("qc") data is available',
        ),
        sa.Column(
            "qc_until",
            sa.TIMESTAMP(),
            nullable=True,
            comment='The timestamp until when quality checked("qc") data is available',
        ),
        sa.Column(
            "qc_droped",
            sa.FLOAT(),
            nullable=True,
            comment="The percentage of droped values during the quality check",
        ),
        sa.PrimaryKeyConstraint("station_id"),
        comment="The Meta informations of the precipitation stations.",
    )
    op.create_index(
        "idx_meta_p_geometry",
        "meta_p",
        ["geometry"],
        unique=False,
        postgresql_using="gist",
    )
    op.create_index(
        "idx_meta_p_geometry_utm",
        "meta_p",
        ["geometry_utm"],
        unique=False,
        postgresql_using="gist",
    )
    op.create_table(
        "meta_p_d",
        sa.Column(
            "station_id",
            sa.Integer(),
            nullable=False,
            comment="official DWD-ID of the station",
        ),
        sa.Column(
            "is_real",
            sa.Boolean(),
            nullable=False,
            comment=" 'Is this station a real station with own measurements or only a virtual station, to have complete timeseries for every precipitation station.",
        ),
        sa.Column(
            "raw_from",
            sa.TIMESTAMP(),
            nullable=True,
            comment='The timestamp from when on own "raw" data is available',
        ),
        sa.Column(
            "raw_until",
            sa.TIMESTAMP(),
            nullable=True,
            comment='The timestamp until when own "raw" data is available',
        ),
        sa.Column(
            "hist_until",
            sa.TIMESTAMP(),
            nullable=True,
            comment='The timestamp until when own "raw" data is available',
        ),
        sa.Column(
            "filled_from",
            sa.TIMESTAMP(),
            nullable=True,
            comment="The timestamp from when on filled data is available",
        ),
        sa.Column(
            "filled_until",
            sa.TIMESTAMP(),
            nullable=True,
            comment="The timestamp until when filled data is available",
        ),
        sa.Column(
            "last_imp_from",
            sa.TIMESTAMP(),
            nullable=True,
            comment="The minimal timestamp of the last import, that might not yet have been treated",
        ),
        sa.Column(
            "last_imp_until",
            sa.TIMESTAMP(),
            nullable=True,
            comment="The maximal timestamp of the last import, that might not yet have been treated",
        ),
        sa.Column(
            "last_imp_filled",
            sa.Boolean(),
            nullable=False,
            server_default="false",
            comment="Got the last import already filled?",
        ),
        sa.Column(
            "stationshoehe",
            sa.Integer(),
            nullable=False,
            comment="The stations height above the ground in meters",
        ),
        sa.Column(
            "stationsname",
            sa.VARCHAR(length=50),
            nullable=False,
            comment="The stations official name as text",
        ),
        sa.Column(
            "bundesland",
            sa.VARCHAR(length=30),
            nullable=False,
            comment="The state the station is located in",
        ),
        sa.Column(
            "geometry",
            geoalchemy2.types.Geometry(
                geometry_type="POINT",
                srid=4326,
                from_text="ST_GeomFromEWKT",
                name="geometry",
                nullable=False,
            ),
            nullable=False,
            comment="The stations location in the WGS84 coordinate reference system (EPSG:4326)",
        ),
        sa.Column(
            "geometry_utm",
            geoalchemy2.types.Geometry(
                geometry_type="POINT",
                srid=25832,
                from_text="ST_GeomFromEWKT",
                name="geometry",
                nullable=False,
            ),
            nullable=False,
            comment="The stations location in the UTM32 coordinate reference system (EPSG:25832)",
        ),
        sa.PrimaryKeyConstraint("station_id"),
        comment="The Meta informations of the daily precipitation stations.",
    )
    op.create_index(
        "idx_meta_p_d_geometry",
        "meta_p_d",
        ["geometry"],
        unique=False,
        postgresql_using="gist",
    )
    op.create_index(
        "idx_meta_p_d_geometry_utm",
        "meta_p_d",
        ["geometry_utm"],
        unique=False,
        postgresql_using="gist",
    )
    op.create_table(
        "meta_t",
        sa.Column(
            "station_id",
            sa.Integer(),
            nullable=False,
            comment="official DWD-ID of the station",
        ),
        sa.Column(
            "is_real",
            sa.Boolean(),
            nullable=False,
            comment=" 'Is this station a real station with own measurements or only a virtual station, to have complete timeseries for every precipitation station.",
        ),
        sa.Column(
            "raw_from",
            sa.TIMESTAMP(),
            nullable=True,
            comment='The timestamp from when on own "raw" data is available',
        ),
        sa.Column(
            "raw_until",
            sa.TIMESTAMP(),
            nullable=True,
            comment='The timestamp until when own "raw" data is available',
        ),
        sa.Column(
            "hist_until",
            sa.TIMESTAMP(),
            nullable=True,
            comment='The timestamp until when own "raw" data is available',
        ),
        sa.Column(
            "filled_from",
            sa.TIMESTAMP(),
            nullable=True,
            comment="The timestamp from when on filled data is available",
        ),
        sa.Column(
            "filled_until",
            sa.TIMESTAMP(),
            nullable=True,
            comment="The timestamp until when filled data is available",
        ),
        sa.Column(
            "last_imp_from",
            sa.TIMESTAMP(),
            nullable=True,
            comment="The minimal timestamp of the last import, that might not yet have been treated",
        ),
        sa.Column(
            "last_imp_until",
            sa.TIMESTAMP(),
            nullable=True,
            comment="The maximal timestamp of the last import, that might not yet have been treated",
        ),
        sa.Column(
            "last_imp_filled",
            sa.Boolean(),
            nullable=False,
            server_default="false",
            comment="Got the last import already filled?",
        ),
        sa.Column(
            "stationshoehe",
            sa.Integer(),
            nullable=False,
            comment="The stations height above the ground in meters",
        ),
        sa.Column(
            "stationsname",
            sa.VARCHAR(length=50),
            nullable=False,
            comment="The stations official name as text",
        ),
        sa.Column(
            "bundesland",
            sa.VARCHAR(length=30),
            nullable=False,
            comment="The state the station is located in",
        ),
        sa.Column(
            "geometry",
            geoalchemy2.types.Geometry(
                geometry_type="POINT",
                srid=4326,
                from_text="ST_GeomFromEWKT",
                name="geometry",
                nullable=False,
            ),
            nullable=False,
            comment="The stations location in the WGS84 coordinate reference system (EPSG:4326)",
        ),
        sa.Column(
            "geometry_utm",
            geoalchemy2.types.Geometry(
                geometry_type="POINT",
                srid=25832,
                from_text="ST_GeomFromEWKT",
                name="geometry",
                nullable=False,
            ),
            nullable=False,
            comment="The stations location in the UTM32 coordinate reference system (EPSG:25832)",
        ),
        sa.Column(
            "last_imp_qc",
            sa.Boolean(),
            nullable=False,
            server_default="false",
            comment="Got the last import already quality checked?",
        ),
        sa.Column(
            "qc_from",
            sa.TIMESTAMP(),
            nullable=True,
            comment='The timestamp from when on quality checked("qc") data is available',
        ),
        sa.Column(
            "qc_until",
            sa.TIMESTAMP(),
            nullable=True,
            comment='The timestamp until when quality checked("qc") data is available',
        ),
        sa.Column(
            "qc_droped",
            sa.FLOAT(),
            nullable=True,
            comment="The percentage of droped values during the quality check",
        ),
        sa.PrimaryKeyConstraint("station_id"),
        comment="The Meta informations of the temperature stations.",
    )
    op.create_index(
        "idx_meta_t_geometry",
        "meta_t",
        ["geometry"],
        unique=False,
        postgresql_using="gist",
    )
    op.create_index(
        "idx_meta_t_geometry_utm",
        "meta_t",
        ["geometry_utm"],
        unique=False,
        postgresql_using="gist",
    )
    op.create_table(
        "needed_download_time",
        sa.Column(
            "timestamp",
            sa.TIMESTAMP(),
            server_default=sa.text("now()"),
            nullable=False,
            comment="The timestamp when the download hapend.",
        ),
        sa.Column(
            "quantity",
            sa.Integer(),
            nullable=False,
            comment="The number of stations that got downloaded",
        ),
        sa.Column(
            "aggregate",
            sa.String(),
            nullable=False,
            comment="The chosen aggregation. e.g. hourly, 10min, daily, ...",
        ),
        sa.Column(
            "timespan",
            sa.Interval(),
            nullable=False,
            comment="The timespan of the downloaded timeseries. e.g. 2 years",
        ),
        sa.Column(
            "zip", sa.Boolean(), nullable=False, comment="Was the download zipped?"
        ),
        sa.Column(
            "pc",
            sa.String(),
            nullable=False,
            comment="The name of the pc that downloaded the timeseries.",
        ),
        sa.Column(
            "duration",
            sa.Interval(),
            nullable=False,
            comment="The needed time to download and create the timeserie",
        ),
        sa.Column(
            "output_size",
            sa.Integer(),
            nullable=False,
            comment="The size of the created output file in bytes",
        ),
        sa.PrimaryKeyConstraint("timestamp"),
        comment="Saves the time needed to save the timeseries. This helps predicting download time",
    )
    op.create_table(
        "parameter_variables",
        sa.Column(
            "parameter",
            sa.String(length=3),
            nullable=False,
            comment="The parameter for which the variables are valid. e.g. n/p_d/t/et.",
        ),
        sa.Column(
            "start_tstp_last_imp",
            sa.TIMESTAMP(),
            nullable=True,
            comment="At what timestamp did the last complete import start. This is then the maximum timestamp for which to expand the timeseries to.",
        ),
        sa.Column(
            "max_tstp_last_imp",
            sa.TIMESTAMP(),
            nullable=True,
            comment="The maximal timestamp of the last imports raw data of all the timeseries",
        ),
        sa.PrimaryKeyConstraint("parameter"),
        comment="This table is there to save specific variables that are nescesary for the functioning of the scripts",
    )
    op.create_table(
        "raw_files",
        sa.Column(
            "parameter",
            sa.VARCHAR(length=3),
            nullable=False,
            comment="The parameter that got downloaded for this file. e.g. t, et, p_d, n",
        ),
        sa.Column(
            "filepath",
            sa.String(),
            nullable=False,
            comment="The filepath on the CDC Server",
        ),
        sa.Column(
            "modtime",
            sa.TIMESTAMP(),
            nullable=False,
            comment="The modification time on the CDC Server of the coresponding file",
        ),
        sa.PrimaryKeyConstraint("parameter", "filepath"),
        comment="The files that got imported from the CDC Server.",
    )
    op.create_table(
        "richter_parameters",
        sa.Column(
            "precipitation_typ",
            sa.Text(),
            nullable=False,
            comment="The type of precipitation. e.g. 'Schnee', 'Regen', ...",
        ),
        sa.Column(
            "e", sa.FLOAT(), nullable=False, comment="The e-value of the equation."
        ),
        sa.Column(
            "b_no-protection",
            sa.FLOAT(),
            nullable=False,
            comment="The b-value of the equation for exposition class 'no protection'.",
        ),
        sa.Column(
            "b_little-protection",
            sa.FLOAT(),
            nullable=False,
            comment="The b-value of the equation for exposition class 'little protection'.",
        ),
        sa.Column(
            "b_protected",
            sa.FLOAT(),
            nullable=False,
            comment="The b-value of the equation for exposition class 'protected'.",
        ),
        sa.Column(
            "b_heavy-protection",
            sa.FLOAT(),
            nullable=False,
            comment="The b-value of the equation for exposition class 'heavy protection'.",
        ),
        sa.PrimaryKeyConstraint("precipitation_typ"),
        comment="The Richter values for the equation.",
    )
    op.create_table(
        "settings",
        sa.Column(
            "key",
            sa.String(length=20),
            nullable=False,
            comment="The key of the setting",
        ),
        sa.Column(
            "value",
            sa.String(length=20),
            nullable=False,
            comment="The value of the setting",
        ),
        sa.PrimaryKeyConstraint("key"),
        comment="This table saves settings values for the script-databse connection. E.G. the latest package version that updated the database.",
    )
    op.create_table(
        "station_ma_raster",
        sa.Column(
            "station_id",
            sa.Integer(),
            nullable=False,
            comment="The DWD-ID of the station.",
        ),
        sa.Column(
            "raster_key",
            sa.VARCHAR(length=7),
            nullable=False,
            comment="The name of the raster. e.g. 'dwd' or 'hyras'",
        ),
        sa.Column(
            "parameter",
            sa.VARCHAR(length=7),
            nullable=False,
            comment="The parameter of the raster. e.g. 'p_wihj', 'p_sohj', 'p_year', 't_year', 'et_year'",
        ),
        sa.Column(
            "value",
            sa.Integer(),
            nullable=False,
            comment="The value of the raster for the station.",
        ),
        sa.Column(
            "distance",
            sa.Integer(),
            nullable=False,
            comment="The distance of the station to the raster value in meters.",
        ),
        sa.PrimaryKeyConstraint("station_id", "raster_key", "parameter"),
        schema="public",
        comment="The multi annual climate raster values for each station.",
    )
    op.create_table(
        "station_ma_timeserie",
        sa.Column(
            "station_id",
            sa.Integer(),
            nullable=False,
            comment="The DWD-ID of the station.",
        ),
        sa.Column(
            "parameter",
            sa.VARCHAR(length=3),
            nullable=False,
            comment="The parameter of the station. e.g. 'p', 't', 'et'",
        ),
        sa.Column(
            "kind",
            sa.String(),
            nullable=False,
            comment="The kind of the timeserie. e.g. 'raw', 'filled', 'corr'",
        ),
        sa.Column(
            "value",
            sa.Float(),
            nullable=False,
            comment="The multi annual value of the yearly mean value of the station to the multi annual mean value of the raster.",
        ),
        sa.PrimaryKeyConstraint("station_id", "parameter", "kind"),
        schema="public",
        comment="The multi annual mean values of the stations timeseries for the maximum available timespan.",
    )


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("station_ma_raster")
    op.drop_table("station_ma_timeserie")
    op.drop_table("settings")
    op.drop_table("richter_parameters")
    op.drop_table("raw_files")
    op.drop_table("parameter_variables")
    op.drop_table("needed_download_time")
    op.drop_index(
        "idx_meta_t_geometry_utm", table_name="meta_t", postgresql_using="gist"
    )
    op.drop_index("idx_meta_t_geometry", table_name="meta_t", postgresql_using="gist")
    op.drop_table("meta_t")
    op.drop_index(
        "idx_meta_p_d_geometry_utm", table_name="meta_p_d", postgresql_using="gist"
    )
    op.drop_index(
        "idx_meta_p_d_geometry", table_name="meta_p_d", postgresql_using="gist"
    )
    op.drop_table("meta_p_d")
    op.drop_index(
        "idx_meta_p_geometry_utm", table_name="meta_p", postgresql_using="gist"
    )
    op.drop_index("idx_meta_p_geometry", table_name="meta_p", postgresql_using="gist")
    op.drop_table("meta_p")
    op.drop_index(
        "idx_meta_et_geometry_utm", table_name="meta_et", postgresql_using="gist"
    )
    op.drop_index("idx_meta_et_geometry", table_name="meta_et", postgresql_using="gist")
    op.drop_table("meta_et")
    op.drop_table("droped_stations")



# Configuration file for the Sphinx documentation builder.

import sys
from pathlib import Path

# -- Path setup --------------------------------------------------------------

src_path = Path(__file__).parent
base_path = src_path.parents[1]
sys.path.insert(0, base_path.resolve().as_posix())

import weatherDB

# -- Project information -----------------------------------------------------

project = 'WeatherDB'
copyright = weatherDB.__copyright__
author = weatherDB.__author__
author_email = weatherDB.__email__

# The full version, including alpha/beta/rc tags
release = weatherDB.__version__


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.napoleon',
    'sphinx_rtd_theme',
    'sphinx.ext.autodoc',
    'nbsphinx',
    'myst_parser',
    'sphinx.ext.viewcode',
    'autoclasstoc',
    'sphinx.ext.autosummary',
    'sphinx_click',
    "sphinx.ext.autosectionlabel",
    'sphinx_copybutton',
    'sphinx_design',
    'sphinx.ext.intersphinx',
    "sphinx_new_tab_link",
    "sphinx.ext.mathjax"
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store', 'sync.ffs_db']

source_suffix = [".rst", ".md"]

source_parsers = {
   '.md': "markdown"
}

myst_enable_extensions = [
    "dollarmath",
    "colon_fence",
    "attrs_inline"
]

# Autodoc options
autodoc_default_options = {
    'member-order': 'bysource',
    'undoc-members': True,
    'show-inheritance': True,
    'inherited-members':True,
    'collapse_navigation': False,
}
autodoc_default_flags=["show-inheritance"]
autoclass_content= "both"
autodoc_inherit_docstrings= True

autoclasstoc_sections = [
    'public-methods',
]

# Autosummary options
autosummary_generate = True
autosummary_generate_overwrite = True

# intersphinx
intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
    'sqlalchemy': ('https://docs.sqlalchemy.org/en/20/', None),
    'pandas': ('https://pandas.pydata.org/docs/', None),}

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "sphinx_rtd_theme"

html_theme_options = {
    'version_selector': True,
    'prev_next_buttons_location': 'bottom',
    'style_external_links': True,
    # TOC options
    'collapse_navigation': True,
    'sticky_navigation': True,
    'navigation_depth': -1,
}

html_css_files = [
    'custom.css',
]
html_logo = "_static/logo.png"
html_favicon = "_static/favicon.ico"

html_context = {
    # display gitlab
    "display_gitlab": True, # Integrate Gitlab
    "gitlab_user": "hydrology", # Username
    "gitlab_repo": "weatherDB", # Repo name
    "gitlab_version": "master", # Version
    "conf_py_path": "/docs/source/", # Path in the checkout to the docs root
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']


# -- Options for PDF Output --------------------------------------------------
latex_engine = "pdflatex"
latex_theme = "manual"


# run with command: sphinx-build -b html .\source .\html


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""A collection of geometry functions.

Based on `max_fun` package on https://github.com/maxschmi/max_fun
Created by Max Schmit, 2021
"""
# libraries
import numpy as np
from shapely.geometry import Point, LineString
import geopandas as gpd
import rasterio as rio

# functions
def polar_line(center_xy, radius, angle):
    """Create a LineString with polar coodinates.

    Parameters
    ----------
    center_xy : list, array or tuple of int or floats
        The X and Y coordinates of the center.
    radius : int or float
        The radius of the circle.
    angle : int
        The angle of the portion of the circle in degrees.
        0 means east.

    Returns
    -------
    shapely.geometry.LineString
        LineString.
    """
    coords = [center_xy]
    coords.append([
            center_xy[0] + np.cos(np.deg2rad(angle)) * radius,
            center_xy[1] + np.sin(np.deg2rad(angle)) * radius
        ])

    return LineString(coords)

def raster2points(raster_np, transform, crs=None):
    """Polygonize raster array to GeoDataFrame.

    Until now this only works for rasters with one band.

    Parameters
    ----------
    raster_np : np.array
        The imported raster array.
    transform : rio.Affine
        The Affine transformation of the raster.
    crs : str or crs-type, optional
        The coordinate reference system for the raster, by default None

    Returns
    -------
    geopandas.GeoDataFrame
        The raster Data is in the data column.
    """
    mask = ~np.isnan(raster_np[0])
    cols, rows =  mask.nonzero()
    coords = rio.transform.xy(transform, cols, rows)

    geoms = [Point(xy) for xy in list(zip(*coords))]

    return gpd.GeoDataFrame(
        {"data": raster_np[0][mask]},
        geometry=geoms,
        crs=crs)


# libraries
import logging

from ..station import StationT
from .StationsBaseTET import StationsBaseTET

# set settings
# ############
__all__ = ["StationsT"]
log = logging.getLogger(__name__)

# class definition
##################

class StationsT(StationsBaseTET):
    """A class to work with and download temperature data for several stations."""
    _StationClass = StationT
    _timeout_raw_imp = 120

    def get_quotient(self, **kwargs):
        raise NotImplementedError("The quotient is not yet implemented for temperature.")



"""
This module has grouping classes for all the stations of one parameter. E.G. StationsP (or StationsP) groups all the Precipitation Stations available.
Those classes can get used to do actions on all the stations.
"""
from .StationsP import StationsP
from .StationsPD import StationsPD
from .StationsT import StationsT
from .StationsET import StationsET

__all__ = ["StationsP", "StationsPD", "StationsT", "StationsET"]


# libraries
import logging
from datetime import datetime
from pathlib import Path
import warnings
import zipfile
from packaging import version
import pandas as pd

from ..utils.TimestampPeriod import TimestampPeriod
from .StationBases import StationBase, AGG_TO
from . import StationP, StationT, StationET

# set settings
# ############
__all__ = ["GroupStation"]
log = logging.getLogger(__name__)

# class definition
##################
class GroupStation(object):
    """A class to group all possible parameters of one station.

    So if you want to create the input files for a simulation, where you need T, ET and N, use this class to download the data for one station.
    """

    def __init__(self, id, error_if_missing=True, **kwargs):
        self.id = id
        self.station_parts = []
        self._error_if_missing = error_if_missing
        for StatClass in [StationP, StationT, StationET]:
            try:
                self.station_parts.append(
                    StatClass(id=id, **kwargs)
                )
            except Exception as e:
                if error_if_missing:
                    raise e
        self.paras_available = [stat._para for stat in self.station_parts]

    def _check_paras(self, paras):
        if isinstance(paras, str) and paras != "all":
            paras = [paras,]

        if isinstance(paras, str) and (paras == "all"):
            return self.paras_available
        else:
            paras_new = []
            for para in paras:
                if para in self.paras_available:
                    paras_new.append(para)
                elif self._error_if_missing:
                    raise ValueError(
                        f"The parameter {para} you asked for is not available for station {self.id}")
            return paras_new

    @staticmethod
    def _check_kinds(kinds):
        # type cast kinds
        if isinstance(kinds, str):
            kinds = [kinds]
        else:
            kinds = kinds.copy()
        return kinds

    def get_available_paras(self, short=False):
        """Get the possible parameters for this station.

        Parameters
        ----------
        short : bool, optional
            Should the short name of the parameters be returned.
            The default is "long".

        Returns
        -------
        list of str
            A list of the long parameter names that are possible for this station to get.
        """
        paras = []
        attr_name = "_para" if short else "_para_long"
        for stat in self.station_parts:
            paras.append(getattr(stat, attr_name))

        return paras

    def get_filled_period(self, kinds="best", from_meta=True, join_how="inner"):
        """Get the combined filled period for all 3 stations.

        This is the maximum possible timerange for these stations.

        Parameters
        ----------
        kind :  str
            The data kind to look for filled period.
            Must be a column in the timeseries DB.
            Must be one of "raw", "qc", "filled", "adj".
            If "best" is given, then depending on the parameter of the station the best kind is selected.
            For Precipitation this is "corr" and for the other this is "filled".
            For the precipitation also "qn" and "corr" are valid.
        from_meta : bool, optional
            Should the period be from the meta table?
            If False: the period is returned from the timeserie. In this case this function is only a wrapper for .get_period_meta.
            The default is True.
        join_how : str, optional
            How should the different periods get joined.
            If "inner" then the minimal period that is inside of all the filled_periods is returned.
            If "outer" then the maximal possible period is returned.
            The default is "inner".

        Returns
        -------
        TimestampPeriod
            The maximum filled period for the 3 parameters for this station.
        """
        kinds = self._check_kinds(kinds)
        for kind in ["filled_by", "adj"]:
            if kind in kinds:
                kinds.remove(kind)

        # get filled_period
        for kind in kinds:
            for stat in self.station_parts:
                new_filled_period =  stat.get_filled_period(
                    kind=kind, from_meta=from_meta)

                if "filled_period" not in locals():
                    filled_period = new_filled_period.copy()
                else:
                    filled_period = filled_period.union(
                        new_filled_period, how=join_how)

        return filled_period

    def get_df(self, period=(None, None), kinds="best", paras="all",
               agg_to="day", nas_allowed=True, add_na_share=False,
               add_t_min=False, add_t_max=False, **kwargs):
        """Get a DataFrame with the corresponding data.

        Parameters
        ----------
        period : TimestampPeriod or (tuple or list of datetime.datetime or None), optional
            The minimum and maximum Timestamp for which to get the timeseries.
            If None is given, the maximum or minimal possible Timestamp is taken.
            The default is (None, None).
        kinds :  str or list of str
            The data kind to look for filled period.
            Must be a column in the timeseries DB.
            Must be one of "raw", "qc", "filled", "adj", "filled_by", "best"("corr" for N and "filled" for T and ET).
            If "best" is given, then depending on the parameter of the station the best kind is selected.
            For Precipitation this is "corr" and for the other this is "filled".
            For the precipitation also "qn" and "corr" are valid.
        agg_to : str, optional
            To what aggregation level should the timeseries get aggregated to.
            The minimum aggregation for Temperatur and ET is daily and for the precipitation it is 10 minutes.
            If a smaller aggregation is selected the minimum possible aggregation for the respective parameter is returned.
            So if 10 minutes is selected, than precipitation is returned in 10 minuets and T and ET as daily.
            The default is "10 min".
        nas_allowed : bool, optional
            Should NAs be allowed?
            If True, then the maximum possible period is returned, even if there are NAs in the timeserie.
            If False, then the minimal filled period is returned.
            The default is True.
        paras : list of str or str, optional
            Give the parameters for which to get the meta information.
            Can be "n", "t", "et" or "all".
            If "all", then every available station parameter is returned.
            The default is "all"
        add_na_share : bool, optional
            Should one or several columns be added to the Dataframe with the share of NAs in the data.
            This is especially important, when the stations data get aggregated, because the aggregation doesn't make sense if there are a lot of NAs in the original data.
            If True, one column per asked kind is added with the respective share of NAs, if the aggregation step is not the smallest.
            The "kind"_na_share column is in percentage.
            The default is False.
        add_t_min : bool, optional
            Should the minimal temperature value get added?
            The default is False.
        add_t_max : bool, optional
            Should the maximal temperature value get added?
            The default is False.

        Returns
        -------
        pd.Dataframe
            A DataFrame with the timeseries for this station and the given period.
        """
        paras = self._check_paras(paras)

        # download dataframes
        dfs = []
        for stat in self.station_parts:
            if stat._para in paras:
                # check if min and max for temperature should get added
                use_kinds = kinds.copy()
                if stat._para == "t":
                    if isinstance(use_kinds, str):
                        use_kinds=[use_kinds]
                    if "best" in use_kinds:
                        use_kinds.insert(use_kinds.index("best"), "filled")
                        use_kinds.remove("best")
                    for k in ["raw", "filled"]:
                        if k in use_kinds:
                            if add_t_max:
                                use_kinds.insert(
                                    use_kinds.index(k)+1,
                                    f"{k}_max")
                            if add_t_min:
                                use_kinds.insert(
                                    use_kinds.index(k)+1,
                                    f"{k}_min")

                # get the data from station object
                df = stat.get_df(
                    period=period,
                    kinds=use_kinds,
                    agg_to=agg_to,
                    nas_allowed=nas_allowed,
                    add_na_share=add_na_share,
                    **kwargs)
                df = df.rename(dict(zip(
                    df.columns,
                    [stat._para.upper() + "_" + col for col in df.columns])),
                    axis=1)
                dfs.append(df)

        # concat the dfs
        if len(dfs) > 1:
            df_all = pd.concat(dfs, axis=1)
        elif len(dfs) == 1 :
            df_all = dfs[0]
        else:
            raise ValueError("No timeserie was found for {paras} and Station {stid}".format(
                paras=", ".join(paras),
                stid=self.id))

        return df_all

    @classmethod
    def get_meta_explanation(cls, infos="all"):
        """Get the explanations of the available meta fields.

        Parameters
        ----------
        infos : list or string, optional
            The infos you wish to get an explanation for.
            If "all" then all the available information get returned.
            The default is "all"

        Returns
        -------
        pd.Series
            a pandas Series with the information names as index and the explanation as values.
        """
        return StationBase.get_meta_explanation(infos=infos)

    def get_max_period(self, kinds, nas_allowed=False):
        """Get the maximum available period for this stations timeseries.

        If nas_allowed is True, then the maximum range of the timeserie is returned.
        Else the minimal filled period is returned

        Parameters
        ----------
        kinds : str or list of str
            The data kinds to update.
            Must be a column in the timeseries DB.
            Must be one of "raw", "qc", "filled", "adj".
            For the precipitation also "qn" and "corr" are valid.
        nas_allowed : bool, optional
            Should NAs be allowed?
            If True, then the maximum possible period is returned, even if there are NAs in the timeserie.
            If False, then the minimal filled period is returned.
            The default is False.

        Returns
        -------
        utils.TimestampPeriod
            The maximum Timestamp Period
        """
        kinds = self._check_kinds(kinds)
        max_period = None
        for stat in self.station_parts:
            max_period_i = stat.get_max_period(
                kinds=kinds, nas_allowed=nas_allowed)
            if max_period is None:
                max_period = max_period_i
            else:
                max_period = max_period.union(
                    max_period_i,
                    how="outer" if nas_allowed else "inner")

        return max_period

    def get_meta(self, paras="all", **kwargs):
        """Get the meta information for every parameter of this station.

        Parameters
        ----------
        paras : list of str or str, optional
            Give the parameters for which to get the meta information.
            Can be "n", "t", "et" or "all".
            If "all", then every available station parameter is returned.
            The default is "all"
        **kwargs : dict, optional
            The optional keyword arguments are handed to the single Station get_meta methods. Can be e.g. "info".

        Returns
        -------
        dict
            dict with the information.
            there is one subdict per parameter.
            If only one parameter is asked for, then there is no subdict, but only a single value.
        """
        paras = self._check_paras(paras)

        for stat in self.station_parts:
            if stat._para in paras:
                meta_para = stat.get_meta(**kwargs)
                if "meta_all" not in locals():
                    meta_all = {stat._para:meta_para}
                else:
                    meta_all.update({stat._para:meta_para})
        return meta_all

    def get_geom(self, crs=None):
        """Get the point geometry of the station.

        Parameters
        ----------
        crs: str, int or None, optional
            The coordinate reference system of the geometry.
            If None, then the geometry is returned in WGS84 (EPSG:4326).
            If string, then it should be in a pyproj readable format.
            If int, then it should be the EPSG code.
            The default is None.

        Returns
        -------
        shapely.geometries.Point
            The location of the station as shapely Point in the given coordinate reference system.
        """
        return self.station_parts[0].get_geom(crs=crs)

    def get_name(self):
        return self.station_parts[0].get_name()

    def create_roger_ts(self, dir, period=(None, None),
                        kind="best", r_r0=1, add_t_min=False, add_t_max=False,
                        do_toolbox_format=False,
                        **kwargs):
        """Create the timeserie files for roger as csv.

        This is only a wrapper function for create_ts with some standard settings.

        Parameters
        ----------
        dir : pathlib like object or zipfile.ZipFile
            The directory or Zipfile to store the timeseries in.
            If a zipfile is given a folder with the statiopns ID is added to the filepath.
        period : TimestampPeriod like object, optional
            The period for which to get the timeseries.
            If (None, None) is entered, then the maximal possible period is computed.
            The default is (None, None)
        kind :  str
            The data kind to look for filled period.
            Must be a column in the timeseries DB.
            Must be one of "raw", "qc", "filled", "adj".
            If "best" is given, then depending on the parameter of the station the best kind is selected.
            For Precipitation this is "corr" and for the other this is "filled".
            For the precipitation also "qn" and "corr" are valid.
        r_r0 : int or float, list of int or float or None, optional
            Should the ET timeserie contain a column with R/R0.
            If None, then no column is added.
            If int or float, then a R/R0 column is appended with this number as standard value.
            If list of int or floats, then the list should have the same length as the ET-timeserie and is appanded to the Timeserie.
            If pd.Series, then the index should be a timestamp index. The serie is then joined to the ET timeserie.
            The default is 1.
        add_t_min=False : bool, optional
            Schould the minimal temperature value get added?
            The default is False.
        add_t_max=False : bool, optional
            Schould the maximal temperature value get added?
            The default is False.
        do_toolbox_format : bool, optional
            Should the timeseries be saved in the RoGeR toolbox format? (have a look at the RoGeR examples in https://github.com/Hydrology-IFH/roger)
            The default is False.
        **kwargs:
            additional parameters for Station.get_df

        Raises
        ------
        Warning
            If there are NAs in the timeseries or the period got changed.
        """
        if do_toolbox_format:
            return self.create_ts(
                dir=dir, period=period, kinds=kind,
                agg_to="10 min", r_r0=r_r0, split_date=True,
                nas_allowed=False,
                add_t_min=add_t_min, add_t_max=add_t_max,
                file_names={"N":"PREC.txt", "T":"TA.txt", "ET":"PET.txt"},
                col_names={"N":"PREC", "ET":"PET",
                           "T":"TA", "T_min":"TA_min", "T_max":"TA_max",
                           "Jahr":"YYYY", "Monat":"MM", "Tag":"DD",
                           "Stunde":"hh", "Minute":"mm"},
                add_meta=False,
                keep_date_parts=True,
                **kwargs)
        else:
            return self.create_ts(
                dir=dir, period=period, kinds=kind,
                agg_to="10 min", r_r0=r_r0, split_date=True,
                nas_allowed=False,
                add_t_min=add_t_min, add_t_max=add_t_max,
                **kwargs)

    def create_ts(self, dir, period=(None, None),
                  kinds="best", paras="all",
                  agg_to="10 min", r_r0=None, split_date=False,
                  nas_allowed=True, add_na_share=False,
                  add_t_min=False, add_t_max=False,
                  add_meta=True, file_names={}, col_names={},
                  keep_date_parts=False,
                  **kwargs):
        """Create the timeserie files as csv.

        Parameters
        ----------
        dir : pathlib like object or zipfile.ZipFile
            The directory or Zipfile to store the timeseries in.
            If a zipfile is given a folder with the statiopns ID is added to the filepath.
        period : TimestampPeriod like object, optional
            The period for which to get the timeseries.
            If (None, None) is entered, then the maximal possible period is computed.
            The default is (None, None)
        kinds :  str or list of str
            The data kinds to look for filled period.
            Must be a column in the timeseries DB.
            Must be one of "raw", "qc", "filled", "adj", "filled_by", "filled_share", "best".
            If "best" is given, then depending on the parameter of the station the best kind is selected.
            For precipitation this is "corr" and for the other this is "filled".
            For the precipitation also "qn" and "corr" are valid.
            If only one kind is asked for, then the columns get renamed to only have the parameter name as column name.
        paras : list of str or str, optional
            Give the parameters for which to get the meta information.
            Can be "n", "t", "et" or "all".
            If "all", then every available station parameter is returned.
            The default is "all"
        agg_to : str, optional
            To what aggregation level should the timeseries get aggregated to.
            The minimum aggregation for Temperatur and ET is daily and for the precipitation it is 10 minutes.
            If a smaller aggregation is selected the minimum possible aggregation for the respective parameter is returned.
            So if 10 minutes is selected, than precipitation is returned in 10 minuets and T and ET as daily.
            The default is "10 min".
        r_r0 : int or float or None or pd.Series or list, optional
            Should the ET timeserie contain a column with R/R0.
            If None, then no column is added.
            If int, then a R/R0 column is appended with this number as standard value.
            If list of int or floats, then the list should have the same length as the ET-timeserie and is appanded to the Timeserie.
            If pd.Series, then the index should be a timestamp index. The serie is then joined to the ET timeserie.
            The default is None.
        split_date : bool, optional
            Should the timestamp get splitted into parts, so one column for year, one for month etc.?
            If False the timestamp is saved in one column as string.
        nas_allowed : bool, optional
            Should NAs be allowed?
            If True, then the maximum possible period is returned, even if there are NAs in the timeserie.
            If False, then the minimal filled period is returned.
            The default is True.
        add_na_share : bool, optional
            Should one or several columns be added to the Dataframe with the share of NAs in the data.
            This is especially important, when the stations data get aggregated, because the aggregation doesn't make sense if there are a lot of NAs in the original data.
            If True, one column per asked kind is added with the respective share of NAs, if the aggregation step is not the smallest.
            The "kind"_na_share column is in percentage.
            The default is False.
        add_t_min=False : bool, optional
            Should the minimal temperature value get added?
            The default is False.
        add_t_max=False : bool, optional
            Should the maximal temperature value get added?
            The default is False.
        add_meta : bool, optional
            Should station Meta information like name and Location (lat, long) be added to the file?
            The default is True.
        file_names : dict, optional
            A dictionary with the file names for the different parameters.
            e.g.{"N":"PREC.txt", "T":"TA.txt", "ET":"ET.txt"}
            If an empty dictionary is given, then the standard names are used.
            The default is {}.
        col_names : dict, optional
            A dictionary with the column names for the different parameters.
            e.g.{"N":"PREC", "T":"TA", "ET":"ET", "Jahr":"YYYY", "Monat":"MM", "Tag":"DD", "Stunde":"HH", "Minute":"MN"}
            If an empty dictionary is given, then the standard names are used.
            The default is {}.
        keep_date_parts : bool, optional
            only used if split_date is True.
            Should the date parts that are not needed, e.g. hour value for daily timeseries, be kept?
            If False, then the columns that are not needed are dropped.
            The default is False.
        **kwargs:
            additional parameters for Station.get_df

        Raises
        ------
        Warning
            If there are NAs in the timeseries and nas_allowed is False
            or the period got changed.
        """
        # check directory
        dir = self._check_dir(dir)

        # type cast kinds
        kinds = self._check_kinds(kinds)
        paras = self._check_paras(paras)

        # get the period
        if not ("_skip_period_check" in kwargs and kwargs["_skip_period_check"]):
            period = TimestampPeriod._check_period(period).expand_to_timestamp()
            period_filled = self.get_filled_period(
                kinds=kinds,
                join_how="outer" if nas_allowed else "inner")

            if period.is_empty():
                period = period_filled
            else:
                period_new = period_filled.union(
                    period,
                    how="inner")
                if period_new != period:
                    warnings.warn(
                        f"The Period for Station {self.id} got changed from {str(period)} to {str(period_new)}.")
                    period = period_new
        if "_skip_period_check" in kwargs:
            del kwargs["_skip_period_check"]

        # prepare loop
        name_suffix = "_{stid:0>5}.txt".format(stid=self.id)
        x, y = self.get_geom().coords.xy
        name = self.get_name() + " (ID: {stid})".format(stid=self.id)
        do_zip = isinstance(dir, zipfile.ZipFile)

        for para in paras:
            # get the timeserie
            df = self.get_df(
                period=period, kinds=kinds,
                paras=[para], agg_to=agg_to,
                nas_allowed=nas_allowed,
                add_na_share=add_na_share,
                add_t_min=add_t_min, add_t_max=add_t_max,
                _skip_period_check=True,
                **kwargs)

            # rename columns
            if len(kinds)==1 or ("filled_by" in kinds and len(kinds)==2):
                if len(kinds)==1:
                    colname_base = [col for col in df.columns if len(col.split("_"))==2][0]
                else:
                    colname_base = f"{para.upper()}_" + kinds[1-(kinds.index("filled_by"))]
                df.rename(
                    {colname_base: para.upper(),
                     f"{colname_base}_min": f"{para.upper()}_min",
                     f"{colname_base}_max": f"{para.upper()}_max",},
                    axis=1, inplace=True)
            else:
                df.rename(
                    dict(zip(df.columns,
                            [col.replace(f"{para}_", f"{para.upper()}_")
                            for col in df.columns])),
                    axis=1, inplace=True)

            # check for NAs
            filled_cols = [col for col in df.columns if "filled_by" in col]
            if not nas_allowed and df.drop(filled_cols, axis=1).isna().sum().sum() > 0:
                warnings.warn("There were NAs in the timeserie for Station {stid}.".format(
                    stid=self.id))

            # special operations for et
            if para == "et" and r_r0 is not None:
                if isinstance(r_r0, int) or isinstance(r_r0, float):
                    df = df.join(
                        pd.Series([r_r0]*len(df), name="R/R0", index=df.index))
                elif isinstance(r_r0, pd.Series):
                    df = df.join(r_r0.rename("R_R0"))
                elif isinstance(r_r0, list):
                    df = df.join(
                        pd.Series(r_r0, name="R/R0", index=df.index))

            # create tables
            if split_date:
                n_parts = 5 if keep_date_parts else AGG_TO[agg_to]["split"][para]
                df = self._split_date(df.index)\
                        .iloc[:, 0:n_parts]\
                        .join(df)
            else:
                df.reset_index(inplace=True)

            # rename columns if user asked for
            df.rename(col_names, axis=1, inplace=True)

            # create header
            if add_meta:
                header = ("Name: " + name + "\t" * (len(df.columns)-1) + "\n" +
                        "Lat: " + y + "   ,Lon: " + x + "\t" * (len(df.columns)-1) + "\n")
            else:
                header = ""

            # get file name
            if para.upper() in file_names:
                file_name = file_names[para.upper()]
            elif para in file_names:
                file_name = file_names[para]
            else:
                file_name = para.upper() + name_suffix

            # write table out
            if version.parse(pd.__version__) > version.parse("1.5.0"):
                to_csv_kwargs = dict(lineterminator="\n")
            else:
                to_csv_kwargs = dict(line_terminator="\n")
            str_df = header + df.to_csv(
                sep="\t", decimal=".", index=False, **to_csv_kwargs)

            if do_zip:
                dir.writestr(f"{self.id}/{file_name}", str_df)
            else:
                with open(dir.joinpath(file_name), "w") as f:
                    f.write(str_df)

    @staticmethod
    def _check_dir(dir):
        """Checks if a directors is valid and empty.

        If not existing the directory is created.

        Parameters
        ----------
        dir : pathlib object or zipfile.ZipFile
            The directory to check.

        Raises
        ------
        ValueError
            If the directory is not empty.
        ValueError
            If the directory is not valid. E.G. it is a file path.
        """
        # check types
        if isinstance(dir, str):
            dir = Path(dir)

        # check directory
        if isinstance(dir, Path):
            if dir.is_dir():
                if len(list(dir.iterdir())) > 0:
                    raise ValueError(
                        "The given directory '{dir}' is not empty.".format(
                            dir=str(dir)))
            elif dir.suffix == "":
                dir.mkdir()
            else:
                raise ValueError(
                    "The given directory '{dir}' is not a directory.".format(
                        dir=dir))
        elif not isinstance(dir, zipfile.ZipFile):
            raise ValueError(
                "The given directory '{dir}' is not a directory or zipfile.".format(
                    dir=dir))

        return dir

    @staticmethod
    def _split_date(dates):
        """
        Split datetime into parts.

        Parameters
        ----------
        dates : pandas.DatetimeIndex or list of (datetime.dateime or pandas.Timestamp) or
                pandas.DataFrame of (datetime.datetime or pandas.Timestamp)
            The datetime's to split.

        Returns
        -------
        pandas.DataFrame
            A DataFrame with 5 columns (Jahr, Monat, Tag, Stunde, Minute).
        """
        # if dates is not a list make it a list
        if isinstance(dates, datetime) or isinstance(dates, pd.Timestamp):
            dates = pd.DatetimeIndex([dates])
            index = range(0, len(dates))

        elif isinstance(dates, pd.DatetimeIndex):
            index = dates
        else:
            index = range(0, len(dates))

        # check if date is datetime or Timestamp:
        if not (isinstance(dates[0], pd.Timestamp) or
                isinstance(dates[0], datetime)):
            raise ValueError("Error: The given date is not in a datetime or " +
                            "Timestamp format.")

        return pd.DataFrame(
            {"Jahr": dates.year,
             "Monat": dates.month,
             "Tag": dates.day,
             "Stunde": dates.hour,
             "Minute": dates.minute},
             dtype=int,
             index=index)



# libraries
import logging

from ..station import StationET
from .StationsBaseTET import StationsBaseTET

# set settings
# ############
__all__ = ["StationsET"]
log = logging.getLogger(__name__)

# class definition
##################
class StationsET(StationsBaseTET):
    """A class to work with and download potential Evapotranspiration (VPGB) data for several stations."""
    _StationClass = StationET
    _timeout_raw_imp = 120



# libraries
import warnings
import pandas as pd
import geopandas as gpd
import logging
import datetime
import socket
import zipfile
from pathlib import Path
from sqlalchemy import text as sqltxt

from ..db.connections import db_engine
from ..utils import TimestampPeriod
from ..station import StationP, GroupStation
from .StationsP import StationsP
from .StationsT import StationsT
from .StationsET import StationsET
from .StationsBase import StationsBase

# set settings
# ############
__all__ = ["GroupStations"]
log = logging.getLogger(__name__)

# class definition
##################

class GroupStations(object):
    """A class to group all possible parameters of all the stations.
    """
    _StationP = StationP
    _GroupStation = GroupStation

    def __init__(self):
        self.stationsN = StationsP()

    def get_valid_stids(self):
        if not hasattr(self, "_valid_stids"):
            sql ="""SELECT station_id FROM meta_p"""
            with db_engine.connect() as con:
                res = con.execute(sqltxt(sql))
            self._valid_stids = [el[0] for el in res.all()]
        return self._valid_stids

    def _check_paras(self, paras):
        if isinstance(paras, str) and paras != "all":
            paras = [paras,]

        valid_paras=["n", "t", "et"]
        if isinstance(paras, str) and (paras == "all"):
            return valid_paras
        else:
            paras_new = []
            for para in paras:
                if para in valid_paras:
                    paras_new.append(para)
                else:
                    raise ValueError(
                        f"The parameter {para} you asked for is not a valid parameter. Please enter one of {valid_paras}")
            return paras_new

    def _check_period(self, period, stids, kinds, nas_allowed=True):
        # get max_period of stations
        max_period = None
        for stid in stids:
            max_period_i = self._GroupStation(stid).get_max_period(
                kinds=kinds, nas_allowed=nas_allowed)
            if max_period is not None:
                max_period = max_period.union(
                    max_period_i,
                    how="outer" if nas_allowed else "inner"
                )
            else:
                max_period=max_period_i

        if not isinstance(period, TimestampPeriod):
            period = TimestampPeriod(*period)
        if period.is_empty():
            return max_period
        else:
            if not period.inside(max_period):
                period = period.union(max_period, how="inner")
                warnings.warn("The asked period is too large. Only {min_tstp} - {max_tstp} is returned".format(
                    **period.get_sql_format_dict(format="%Y-%m-%d %H:%M")))
            return period

    def _check_stids(self, stids):
        """Check if the given stids are valid Station IDs.

        It checks against the Precipitation stations.
        """
        if isinstance(stids, str) and (stids == "all"):
            return self.get_valid_stids()
        else:
            valid_stids = self.get_valid_stids()
            mask_stids_valid = [stid in valid_stids for stid in stids]
            if all(mask_stids_valid):
                return stids
            else:
                raise ValueError(
                    "There is no station defined in the database for the IDs: {stids}".format(
                        stids=", ".join(
                            [str(stid) for stid, valid in zip(stids, mask_stids_valid)
                                  if not valid])))

    @staticmethod
    def _check_dir(dir):
        """Checks if a directors is valid and empty.

        If not existing the directory is created.

        Parameters
        ----------
        dir : pathlib object or zipfile.ZipFile
            The directory to check.

        Raises
        ------
        ValueError
            If the directory is not empty.
        ValueError
            If the directory is not valid. E.G. it is a file path.
        """
        # check types
        if isinstance(dir, str):
            dir = Path(dir)

        # check directory
        if isinstance(dir, zipfile.ZipFile):
            return dir
        elif isinstance(dir, Path):
            if dir.is_dir():
                if len(list(dir.iterdir())) > 0:
                    raise ValueError(
                        "The given directory '{dir}' is not empty.".format(
                            dir=str(dir)))
            elif dir.suffix == "":
                dir.mkdir()
            elif dir.suffix == ".zip":
                if not dir.parent.is_dir():
                    raise ValueError(
                        "The given parent directory '{dir}' of the zipfile is not a directory.".format(
                            dir=dir.parents))
            else:
                raise ValueError(
                    "The given directory '{dir}' is not a directory.".format(
                        dir=dir))
        else:
            raise ValueError(
                "The given directory '{dir}' is not a directory or zipfile.".format(
                    dir=dir))

        return dir

    @classmethod
    def get_meta_explanation(cls, infos="all"):
        """Get the explanations of the available meta fields.

        Parameters
        ----------
        infos : list or string, optional
            The infos you wish to get an explanation for.
            If "all" then all the available information get returned.
            The default is "all"

        Returns
        -------
        pd.Series
            a pandas Series with the information names as index and the explanation as values.
        """
        return cls._GroupStation.get_meta_explanation(infos=infos)

    def get_meta(self, paras="all", stids="all", **kwargs):
        """Get the meta Dataframe from the Database.

        Parameters
        ----------
        paras : list or str, optional
            The parameters for which to get the information.
            If "all" then all the available parameters are requested.
            The default is "all".
        stids: string or list of int, optional
            The Stations to return the meta information for.
            Can either be "all", for all possible stations
            or a list with the Station IDs.
            The default is "all".
        **kwargs: dict, optional
            The keyword arguments are passed to the station.GroupStation().get_meta method.
            From there it is passed to the single station get_meta method.
            Can be e.g. "infos"

        Returns
        -------
        dict of pandas.DataFrame or geopandas.GeoDataFrame
        or pandas.DataFrame or geopandas.GeoDataFrame
            The meta DataFrame.
            If several parameters are asked for, then a dict with an entry per parameter is returned.

        Raises
        ------
        ValueError
            If the given stids (Station_IDs) are not all valid.
        ValueError
            If the given paras are not all valid.
        """
        paras = self._check_paras(paras)
        stats = self.get_para_stations(paras=paras)

        for stat in stats:
            meta_para = stat.get_meta(stids=stids, **kwargs)
            meta_para["parameter"] = stat._para
            if "meta_all" not in locals():
                meta_all = meta_para
            else:
                meta_all = pd.concat([meta_all, meta_para], axis=0)
                if isinstance(meta_para, gpd.GeoDataFrame):
                    meta_all = gpd.GeoDataFrame(meta_all, crs=meta_para.crs)

        if len(paras)==1:
            return meta_all.drop("parameter", axis=1)
        else:
            return meta_all.reset_index().set_index(["station_id", "parameter"]).sort_index()

    def get_para_stations(self, paras="all"):
        """Get a list with all the multi parameter stations as stations.Station*Parameter*-objects.

        Parameters
        ----------
        paras : list or str, optional
            The parameters for which to get the objects.
            If "all" then all the available parameters are requested.
            The default is "all".

        Returns
        -------
        Station-object
            returns a list with the corresponding station objects.

        Raises
        ------
        ValueError
            If the given stids (Station_IDs) are not all valid.
        """
        paras = self._check_paras(paras)
        if not hasattr(self, "stations"):
            self.stations = [self.stationsN, StationsT(), StationsET()]
        return [stats for stats in self.stations if stats._para in paras]

    def get_group_stations(self, stids="all", **kwargs):
        """Get a list with all the stations as station.GroupStation-objects.

        Parameters
        ----------
        stids: string or list of int, optional
            The Stations to return.
            Can either be "all", for all possible stations
            or a list with the Station IDs.
            The default is "all".
        **kwargs: optional
            The keyword arguments are handed to the creation of the single GroupStation objects.
            Can be e.g. "error_if_missing".

        Returns
        -------
        Station-object
            returns a list with the corresponding station objects.

        Raises
        ------
        ValueError
            If the given stids (Station_IDs) are not all valid.
        """
        if "error_if_missing" not in kwargs:
            kwargs.update({"error_if_missing": False})
        kwargs.update({"_skip_meta_check":True})
        valid_stids = self.get_valid_stids()

        if isinstance(stids, str) and (stids == "all"):
            stations = [
                self._GroupStation(stid, **kwargs)
                for stid in valid_stids]
        else:
            stids = list(stids)
            stations = [self._GroupStation(stid, **kwargs)
                        for stid in valid_stids if stid in stids]
            stations_ids = [stat.id for stat in stations]
            if len(stations) != len(stids):
                raise ValueError(
                    "It was not possible to create a {para_long} Station with the following IDs: {stids}".format(
                        para_long=self._para_long,
                        stids = ", ".join([stid for stid in stids if stid in stations_ids])
                    ))

        return stations

    def create_ts(self, dir, period=(None, None), kinds="best",
                  stids="all", agg_to="10 min", r_r0=None, split_date=False,
                  nas_allowed=True, add_na_share=False,
                  add_t_min=False, add_t_max=False,
                  **kwargs):
        """Download and create the weather tables as csv files.

        Parameters
        ----------
        dir : path-like object
            The directory where to save the tables.
            If the directory is a ZipFile, then the output will get zipped into this.
        period : TimestampPeriod like object, optional
            The period for which to get the timeseries.
            If (None, None) is entered, then the maximal possible period is computed.
            The default is (None, None)
        kinds :  str or list of str
            The data kind to look for filled period.
            Must be a column in the timeseries DB.
            Must be one of "raw", "qc", "filled", "adj".
            If "best" is given, then depending on the parameter of the station the best kind is selected.
            For Precipitation this is "corr" and for the other this is "filled".
            For the precipitation also "qn" and "corr" are valid.
        stids : string or list of int, optional
            The Stations for which to compute.
            Can either be "all", for all possible stations
            or a list with the Station IDs.
            The default is "all".
        agg_to : str, optional
            To what aggregation level should the timeseries get aggregated to.
            The minimum aggregation for Temperatur and ET is daily and for the precipitation it is 10 minutes.
            If a smaller aggregation is selected the minimum possible aggregation for the respective parameter is returned.
            So if 10 minutes is selected, than precipitation is returned in 10 minuets and T and ET as daily.
            The default is "10 min".
        r_r0 : int or float or None or pd.Series or list, optional
            Should the ET timeserie contain a column with R/R0.
            If None, then no column is added.
            If int, then a R/R0 column is appended with this number as standard value.
            If list of int or floats, then the list should have the same length as the ET-timeserie and is appended to the Timeserie.
            If pd.Series, then the index should be a timestamp index. The series is then joined to the ET timeserie.
            The default is None.
        split_date : bool, optional
            Should the timestamp get splitted into parts, so one column for year, one for month etc.?
            If False the timestamp is saved in one column as string.
        nas_allowed : bool, optional
            Should NAs be allowed?
            If True, then the maximum possible period is returned, even if there are NAs in the timeserie.
            If False, then the minimal filled period is returned.
            The default is True.
        add_na_share : bool, optional
            Should one or several columns be added to the Dataframe with the share of NAs in the data.
            This is especially important, when the stations data get aggregated, because the aggregation doesn't make sense if there are a lot of NAs in the original data.
            If True, one column per asked kind is added with the respective share of NAs, if the aggregation step is not the smallest.
            The "kind"_na_share column is in percentage.
            The default is False.
        add_t_min : bool, optional
            Should the minimal temperature value get added?
            The default is False.
        add_t_max : bool, optional
            Should the maximal temperature value get added?
            The default is False.
        **kwargs:
            additional parameters for GroupStation.create_ts
        """
        start_time = datetime.datetime.now()
        # check directory and stids
        dir = self._check_dir(dir)
        stids = self._check_stids(stids)

        # check period
        period = self._check_period(
            period=period, stids=stids, kinds=kinds,
            nas_allowed=nas_allowed)
        if period.is_empty():
            raise ValueError("For the given settings, no timeseries could get extracted from the database.\nMaybe try to change the nas_allowed parameter to True, to see, where the problem comes from.")

        # create GroupStation instances
        gstats = self.get_group_stations(stids=stids)
        pbar = StationsBase._get_progressbar(
            max_value=len(gstats),
            name="create TS")
        pbar.update(0)

        if dir.suffix == ".zip":
            with zipfile.ZipFile(
                    dir, "w",
                    compression=zipfile.ZIP_DEFLATED,
                    compresslevel=5) as zf:
                for stat in gstats:
                    stat.create_ts(
                        dir=zf,
                        period=period,
                        kinds=kinds,
                        agg_to=agg_to,
                        r_r0=r_r0,
                        split_date=split_date,
                        nas_allowed=nas_allowed,
                        add_na_share=add_na_share,
                        add_t_min=add_t_min,
                        add_t_max=add_t_max,
                        _skip_period_check=True,
                        **kwargs)
                    pbar.variables["last_station"] = stat.id
                    pbar.update(pbar.value + 1)
        else:
            for stat in gstats:
                stat.create_ts(
                    dir=dir.joinpath(str(stat.id)),
                    period=period,
                    kinds=kinds,
                    agg_to=agg_to,
                    r_r0=r_r0,
                    split_date=split_date,
                    nas_allowed=nas_allowed,
                    add_na_share=add_na_share,
                    add_t_min=add_t_min,
                    add_t_max=add_t_max,
                    **kwargs)
                pbar.variables["last_station"] = stat.id
                pbar.update(pbar.value + 1)

        # get size of output file
        if dir.suffix == ".zip":
            out_size = dir.stat().st_size
        else:
            out_size = sum(
                f.stat().st_size for f in dir.glob('**/*') if f.is_file())

        # save needed time to db
        if db_engine.is_superuser:
            sql_save_time = """
                INSERT INTO needed_download_time(timestamp, quantity, aggregate, timespan, zip, pc, duration, output_size)
                VALUES (now(), '{quantity}', '{agg_to}', '{timespan}', '{zip}', '{pc}', '{duration}', '{out_size}');
            """.format(
                quantity=len(stids),
                agg_to=agg_to,
                timespan=str(period.get_interval()),
                duration=str(datetime.datetime.now() - start_time),
                zip="true" if dir.suffix ==".zip" else "false",
                pc=socket.gethostname(),
                out_size=out_size)
            with db_engine.connect() as con:
                con.execute(sqltxt(sql_save_time))
                con.commit()

        # create log message
        log.debug(
            "The timeseries tables for {quantity} stations got created in {dir}".format(
                quantity=len(stids), dir=dir))

    def create_roger_ts(self, dir, period=(None, None), stids="all",
                        kind="best", r_r0=1,
                        add_t_min=False, add_t_max=False,
                        do_toolbox_format=False, **kwargs):
        """Create the timeserie files for roger as csv.

        This is only a wrapper function for create_ts with some standard settings.

        Parameters
        ----------
        dir : pathlib like object or zipfile.ZipFile
            The directory or Zipfile to store the timeseries in.
            If a zipfile is given a folder with the stations ID is added to the filepath.
        period : TimestampPeriod like object, optional
            The period for which to get the timeseries.
            If (None, None) is entered, then the maximal possible period is computed.
            The default is (None, None)
        stids: string or list of int, optional
            The Stations for which to compute.
            Can either be "all", for all possible stations
            or a list with the Station IDs.
            The default is "all".
        kind :  str
            The data kind to look for filled period.
            Must be a column in the timeseries DB.
            Must be one of "raw", "qc", "filled", "adj".
            If "best" is given, then depending on the parameter of the station the best kind is selected.
            For Precipitation this is "corr" and for the other this is "filled".
            For the precipitation also "qn" and "corr" are valid.
        r_r0 : int or float or None or pd.Series or list, optional
            Should the ET timeserie contain a column with R_R0.
            If None, then no column is added.
            If int, then a R/R0 column is appended with this number as standard value.
            If list of int or floats, then the list should have the same length as the ET-timeserie and is appended to the Timeserie.
            If pd.Series, then the index should be a timestamp index. The series is then joined to the ET timeserie.
            The default is 1.
        add_t_min : bool, optional
            Should the minimal temperature value get added?
            The default is False.
        add_t_max : bool, optional
            Should the maximal temperature value get added?
            The default is False.
        do_toolbox_format : bool, optional
            Should the timeseries be saved in the RoGeR toolbox format? (have a look at the RoGeR examples in https://github.com/Hydrology-IFH/roger)
            The default is False.
        **kwargs:
            additional parameters for GroupStation.create_ts

        Raises
        ------
        Warning
            If there are NAs in the timeseries or the period got changed.
        """
        if do_toolbox_format:
            return self.create_ts(
                dir=dir, period=period, kinds=kind,
                agg_to="10 min", r_r0=r_r0, stids=stids,
                split_date=True, nas_allowed=False,
                add_t_min=add_t_min, add_t_max=add_t_max,
                file_names={"N":"PREC.txt", "T":"TA.txt", "ET":"PET.txt"},
                col_names={"N":"PREC", "ET":"PET",
                           "T":"TA", "T_min":"TA_min", "T_max":"TA_max",
                           "Jahr":"YYYY", "Monat":"MM", "Tag":"DD",
                           "Stunde":"hh", "Minute":"mm"},
                add_meta=False,
                keep_date_parts=True,
                **kwargs)
        else:
            return self.create_ts(
                dir=dir, period=period, kinds=kind,
                agg_to="10 min", r_r0=r_r0, stids=stids,
                split_date=True, nas_allowed=False,
                add_t_min=add_t_min, add_t_max=add_t_max,
                **kwargs)
